#include "CARG_f.h"

CARG_f::CARG_f()
{
  edges  = NULL;
  chr = NULL;
  pos = NULL;
  logprob = NULL;

  mne = 0;
  root = -1;
  nsample = 0;
  exist = 0;
}




CARG_f :: ~CARG_f()
{
  //  printf("coming delete farg\n");
  //  return;
  if (edges != NULL) free(edges);
  if (chr != NULL) free(chr);
  if (pos != NULL) free(pos);
  for (int i = 0; i != nsample; ++i)       free(alnnames[i]);
}



void CARG_f :: Init(int nedges, int nmarker, int rchr, int * marker)
{
  m = nmarker;
  chr = (int*)calloc(m+1, sizeof(int));
  pos = (int*)calloc(m+1, sizeof(int));
  
  for (int i = 0; i != nmarker; ++i) pos[i] = marker[i];
  for (int i = 0; i != nmarker; ++i) chr[i] = rchr;

  mne = nedges;
  edges = (Edge*) malloc(sizeof(Edge) * mne);      
  for (int i = 0; i != mne; ++i) edges[i].nc = 0;
  root = 0;
  nleaf = -1;
}



void CARG_f :: Init (int rchr, int nmarker, int *marker, int rnsample, char ** rnames)
{
  m = nmarker;
  chr = (int*)calloc(m, sizeof(int));
  pos = (int*)calloc(m, sizeof(int));

  for (int i = 0; i != nmarker; ++i) pos[i] = marker[i];
  for (int i = 0; i != nmarker; ++i) chr[i] = rchr;

  nsample = rnsample;

  mne = K10; 
  edges = (Edge*) malloc(sizeof(Edge) * mne);      
  for (int i = 0; i != mne; ++i) edges[i].nc = 0;


  root = 0;
  nleaf = -1;

  nleaf = 2 * nsample;
  for (int i = 0;  i != nsample; ++i)
    {
      alnnames[i] = (char*)calloc(strlen(rnames[i])+1, sizeof(char));
      strcpy(alnnames[i], rnames[i]);
    }
}


void CARG_f :: Add(int id1, int id2, int st, int ed)
{
  while (id1 >= mne)
    {
      int omne = mne;
      mne = id1 + 10000;
      edges = (Edge*) realloc(edges, sizeof(Edge) * mne);
      for (int i = omne; i != mne; ++i) edges[i].nc = 0;
    }

  //
  //  printf("ok here in add: id1 = %d, edges[id1] = %d, mne \n", id1, edges[id1].nc);

  int id = edges[id1].nc++;


  edges[id1].c[id]  = id2;
  edges[id1].st[id] = st;
  edges[id1].ed[id] = ed + 1;



  if (root < id1) root = id1;
  if (root < id2) root = id2;
}






int CARG_f::Init(char *fn, char *fnsample)
{
  if (strlen(fnsample) == 0) return 0;
  printf("Read fast arg trees from file.\n");
  return 0;
}



void CARG_f::Display(int mid)
{
  for (int i =0; i <= root; ++i)
    for (int j = 0; j != edges[i].nc; ++j)
      //      if (edges[i].st[j] <= mid && edges[i].ed[j] > mid)
      if (SAT(i, j, mid))
	printf("edge\t%d\t%d\t%d\t%d\n", i, edges[i].c[j], edges[i].st[j], edges[i].ed[j]);
}

void CARG_f::Display(void)
{
  printf("coming here in display tree\n");

  if (!ExistData()) return;

  printf("------------------ nsample = %d ----\n", nsample);
  for (int i = 0; i != nsample; ++i) 
    printf("farg name = %s\n", alnnames[i]);

  for (int i =0; i != m; ++i)
    printf("marker\t%d\t%d\n", chr[i], pos[i]);


  for (int i =0; i <= root; ++i)
    for (int j = 0; j != edges[i].nc; ++j)
      printf("edge\t%d\t%d\t%d\t%d\n", i, edges[i].c[j], edges[i].st[j], edges[i].ed[j]);

  printf("nleaf (disp) = %d\n", nleaf);
}


int CARG_f::CountChildren(int id, int mid, int &sid)
{
  int nc = 0;
  for (int j = 0; j != edges[id].nc; ++j)    
    //    if (edges[id].st[j] <= mid && mid < edges[id].ed[j] ) 
    if (SAT(id, j, mid))      { nc ++; sid = j;}
  return nc;
}

int CARG_f :: SAT(int id, int j, int mid)
{
  if (edges[id].nc <= j) return 0;
  if (edges[id].st[j] <= mid && mid < edges[id].ed[j]) return 1;
  return 0;
}


int CARG_f::MoveDown(int id, int mid)
{
  do    {
    if (id < nleaf) return id;
    if (edges[id].active == 0) return -1;
    if (edges[id].active == 2) return id; //find the nodes

    for (int j = 0; j != edges[id].nc; ++j)
      if (SAT (id, j, mid))
      {
	  int id1 = edges[id].c[j];
	  if (edges[id1].active > 0) { id = id1; break;}
	}
  } while (id >= nleaf);
  return id;
}


void CARG_f::BuildTree(int r, int mid, int*p, int &nn)  
{
  int c1, c2;
  do {
    if (edges[r].active != 2) {
      printf("errror : %d\n", r);
    }
    c1 = MoveDown(edges[r].c[0], mid);
    c2 = MoveDown(edges[r].c[1], mid);
    if (c1 == -1 && c2 == -1)
      {
	printf("error here: %d %d: %d [mid = %d]\n", c1, c2, r, mid);
	exit(1);
      }

    if (c1 == -1)  r = c2; 
    if (c2 == -1)  r = c1;
  } while (c1 == -1 || c2 == -1);


  p[nn++] = r;
  p[nn++] = c1;
  p[nn++] = c2;


  if (c1 >= nleaf) BuildTree(c1, mid, p, nn);
  if (c2 >= nleaf) BuildTree(c2, mid, p, nn);

}


void CARG_f::DisplayTree(int *p, int nn, int r)
{
  char s [M1];
  strcpy(s, "");
  DisplayTree(p, nn, r, s);
  printf("%s;\n", s);
}

void CARG_f::DisplayTree(int *p, int nn, int r, char *stree)
{
  if (r < nleaf) {
    char tempc[128];
    sprintf(tempc, "%d",  r);
    strcat(stree, tempc);
  }
  else
    {
      for (int i = 0; i != nn; ++i)
	if (p[i] == r && i % 3 == 2)  //root node
	  {
	    strcat(stree, "(");
	    DisplayTree(p, nn,  p[i-2], stree);

	    strcat(stree, ",");
	    DisplayTree(p, nn,  p[i-1], stree);
	    strcat(stree, ")");
	    return;
	  }

      printf("error; %d\n", r);
      exit(1);
    }
}



void CARG_f::VertString2Array(char *s, int *p, int &nn)
{
  int stn = K10; //99999

  int t[K10];
  nn = 0;
  char tempc[128];
  while (1==1)
    {
      int rt = 0;
      int id1, id2;
      while (s[rt] != ')' && s[rt] != ';') rt++;
      if (s[rt] == ';') break;

      int lt = rt;
      while (s[lt] != '(') lt --;

      sscanf(&s[lt+1], "%d", &id1);
      char *p = strstr(&s[lt+1], ",") + 1;
      sscanf(p, "%d", &id2);


      t[nn++] = stn;
      t[nn++] = id1;
      t[nn++] = id2;

      for (int i = lt; i <= rt; ++i) s[i] = ' ';
      sprintf(tempc, "%d", stn);

      int len = strlen(tempc);
      for (int i = 0; i != len; ++i) 
	s[i+lt] = (int) tempc[i];
      stn ++;


      len  = strlen(s);
      int  l = 0;
      for (int i =0; i != len; ++i)
	if (s[i] != ' ') s[l++] = s[i];
      s[l] = '\0';

      //      printf("%s\n", s);
    }

  int nleaf = 0;
  for (int i =0; i != nn; ++i)
    if (t[i] < K10 && t[i] + 1 > nleaf) nleaf = t[i] + 1;


  int down = K10 - nleaf;

  for (int i =0; i != nn; ++i)
    if (t[i] > K10) t[i] -= down;

}


int CARG_f::CheckTree(int *p, int nn, int rp) //rp: 0 (r, ch1, chr) rp =2 (ch1, ch2, r)
{
  int nc = 0;
  if (nc > 1) 
    {
      for (int i = 0; i != nn; ++i) {
	printf("%d\t", p[i]);
	if (i %3 == 2) printf(";");
      }
      printf("\n");
      exit(1);
    }
  return 1;
}


int CARG_f::GetTree(int mid, int*t)
{
  for (int i = 0; i != root + 1; ++i) edges[i].active = -10; //no active
  for (int i = 0; i != nleaf; ++i) edges[i].active = 2; //keep

  for (int i = nleaf; i != root + 1; ++i)
    {
      int nc = 0;
      for (int j = 0; j != edges[i].nc; ++j)
	if (SAT(i, j, mid))
	{
	  int k = edges[i].c[j];
	  if (edges[k].active > 0) nc ++;
	}
      edges[i].active = nc;
    }

  int p[K10];
  int r  = MoveDown(root, mid);
  int nn = 0;
  BuildTree(r, mid, p, nn);

  ReMap(p, nn, t);
  for (int i = 0; i != nn/3; ++i)
    {
      int temp = t[i*3];
      t[i*3] = t[i*3+2];
      t[i*3+2] = temp;
    }
  return nn;
}


void CARG_f::ReMap(int *p, int nn, int *t)
{
  int maxn = 0;
  for (int i = 0; i != nn; ++i) 
    if (maxn  < p[i]) maxn = p[i];


  int* code = (int*)calloc(maxn+10, sizeof(int));
  for (int i = 0; i != maxn+1; ++i) code[i] = -1;
  for (int i = 0; i != nleaf; ++i) code[i] = i;
  
  int nc = nleaf, con = 1;
  while (con)
    {
      con = 0;

      for (int i = 0; i < nn; i += 3)
	if (code[ p[i] ] == -1)
	  if (code[ p[i+1] ] > -1 && code[ p[i+2] ] > -1)
	    { 
	      code[ p[i] ] = nc++; 
	      con = 1;
	    }
    }

  for (int i = 0; i != nn; ++i)    t[i] = code[ p[i] ];
  free(code);
}


int CARG_f::Seek(int rchr, int rpos)
{
  int st = 0;
  int ed = m;
  while (st + 1 < ed)
    {
      int mid = (st + ed)/2;
      if (pos[mid] == rpos) return mid; //find the solution
      if (pos[mid] > rpos) ed = mid;
      else st = mid;
    }
  return st;
}



int CARG_f::GetTrees(int rchr, int rpos, CTree **T)
{
  if (rchr  != chr[0]) return 0; //not in the same chromosome

  int st = -1, ed = -1;

  if (rpos < pos[0])   {st = 0;  ed = 0;}
  if (rpos > pos[m-1]) {st = m-1; ed = m-1;}

  if (st == -1)
    {
      st = Seek(rchr, rpos);
      if (pos[st] == rpos) ed = st;
      else ed = st + 1;
    }

  int t[K10];
  int nleaf = -1, nt = 0;

  //  printf("st = %d, ed = %d\n", st, ed);

  for (int i = st; i != ed + 1; ++i)
    {
      int nn = GetTree(i, t);

      if (nn == 0)
	{
	  printf("error: %d\n", st);
	  exit(1);
	}
      
      if (nleaf == -1) nleaf = (nn/3) + 1;

      if (nleaf != nn/3 + 1)
	{
	  printf("error: %d\t%d\n", chr[i], pos[i]);
	  exit(1);
	}
      tree[i-st].Init(t, nleaf, chr[i], pos[i]);
      T[nt++] = &tree[i-st];
    }


    if (logprob != NULL)
      for (int i =0; i != nt; ++i);
      //    T[i]->InitLikelihoodData(index, logprob);

  return nt;
}




//find the order of nodes in ARG tree and in likelihood samples
void CARG_f :: InitLikelihoodName(char ** lknames, int nslk)
{
  Match(alnnames, nsample, lknames, nslk, index);
}




int CARG_f :: GetNumSamples(void) { return (nleaf / 2);  }

