#include "CARG_m.h"

CARG_m::CARG_m()
{
  maxtree = 0;
  Ts = NULL; //(CTree**) calloc(maxtree, sizeof(CTree*));
  nt = 0;
  stntid = 0;
  logprob = NULL;
  nsp = 0;
  f = NULL;

  bzgip = 0; //using plain text file
  bgzipf = 0;
}

CARG_m::~CARG_m()
{
  printf("nt =%d\n", nt);
  
  for (int i = 0; i != nt; ++i) delete Ts[i];
  free(Ts);
  for (int i = 0; i != nsp; ++i) free(alnnames[i]);

  if (f != NULL) fclose(f);

  if (bgzipf != 0)
    {
      ti_index_destroy(bgzipf_id);
      bgzf_close(bgzipf);
    }
}


void CARG_m::NewTree(void) 
{

  if (nt == maxtree) 
    {
      maxtree += K10;
      Ts = (CTree**) realloc(Ts, maxtree*sizeof(CTree*));
    }
  Ts[nt] = new CTree;
}


void CARG_m :: Reset(void)
{
  for (int i = 0;  i != nt; ++i) delete Ts[i];
  nt = 0;
  stntid = 0;
  fseek (f, 0, SEEK_SET);
}


void CARG_m::Remove(int nrmt)
{
  for (int i = 0; i != nrmt; ++i)        delete Ts[i];
  int l = 0;
  for (int i = nrmt; i != nt; ++i) Ts[l++] = Ts[i];
  stntid = 0;
  nt = l;
}

void CARG_m::ReadAlignName(char *fn)
{
  if (strlen(fn) == 0)      return;

  FILE *fin = fopen(fn, "r");
  if (fin == NULL)
    {
      printf("error file input file for ARGs (MARGARITA format): %s\n", fn);
      exit(1);
    }
  //  printf("fn = %s\n", fn);

  char NAname[1028];             
  nsp = 0;
  while (!feof(fin))
    {
      if (fscanf(fin, "%s", NAname) != 1) continue; 
             
      char *p = NAname;
      if (strlen(p) > 7) 
	printf("warning: name with length > 7,  [%s]\n", p);

      alnnames[nsp] = (char*)calloc(strlen(p) + 1, sizeof(char));               
      strcpy(alnnames[nsp], p);
      nsp ++; 
     if (nsp >= MAXSAMPLE)
	{
	  printf("Check the ARG trees: we found %d samples\n", MAXSAMPLE);
	  exit(1);
	}
    }
  fclose(fin);
}

int  CARG_m::Init(char* fnListSample, char *fnARG, int vbgzip)
{
  bzgip = vbgzip;

  if (strlen(fnARG) == 0)
    {
      printf("No Margarita ARG input\n");
      return 0;
    }

  stntid = 0;
  nt = 0;
  ReadAlignName(fnListSample);
  nleaf = nsp * 2;


  if (vbgzip)
    {
      bgzipf = bgzf_open(fnARG, "r");
      if (bgzipf == 0)
	{
	  fprintf(stderr, "fail to open the data file %s\n", fnARG);
	  return 1;
	}

      bgzipf_id = ti_index_load(fnARG);
      if (bgzipf_id == 0)
	{
	  bgzf_close(bgzipf);
	  fprintf(stderr, "failed to load the index of %s.\n", fnARG);
	  return 1;
	}
     
      printf("init file\n");
 
    }
  else
    {
      f = fopen(fnARG, "r");
      if (f == NULL)
	printf("error input file ARGs (Margarita format): %s\n", fnARG);
    }
  return 1;
}





int  CARG_m::Init(char* fnListSample, char *fnARG)
{
  return Init(fnListSample, fnARG, 0); 
}


//be very careful when using ReadNextTree because nt (the number of tree) would change
int CARG_m::ReadNextTree(int nrt) //end of file 
{
  if (stntid > K1) Remove(stntid-1); //remove the first trees
  if (feof(f)) return 1; //end of file

  //  while (1==1)
  for (int i = 0; i != nrt; ++i)
    {
      //      Ts[nt] = new CTree;
      NewTree();

      if (Ts[nt] == NULL)
	{
	  printf("calloc error in ReadNextTree \n");
	  exit(1);
	}

      if (!Ts[nt]->Init(f)) { delete Ts[nt]; return 1;}

      if (Ts[nt]->GetNumLeaf() != nleaf) {
	printf("error: %d\t%d\n", Ts[nt]->GetChr(), Ts[nt]->GetPos());
	printf("nleaf = %d ntreeleaf = %d\n", nleaf, Ts[nt]->GetNumLeaf());
      }
      nt ++;
    }
  return 0;
}


int CARG_m :: GetTree(int id,  CTree **T)   
{
  int ntree = 0;
  for (int i = id; i != nt; ++i)
    if (Ts[i]->Equal(Ts[id]->GetChr(), Ts[id]->GetPos())) T[ntree++] = Ts[i];
  return ntree;
}


int CARG_m::leftequal(int i)
{
  int j = i;
  while (j > 0)
    if (Ts[j-1]->Equal(Ts[i]->GetChr(), Ts[i]->GetPos())) j --;
    else return j;

  return j;
}




int CARG_m::GetTrees(int chr, int pos, CTree **T)
{
  int nid;
  while (1==1)
    {
      if (nt - stntid  < K1)
	if (bzgip) ReadNextTree_bgzip(chr, pos);
	else  ReadNextTree(K1); 

      nid = -1;
      for (int i = stntid; i != nt; ++i)
	if (!Ts[stntid]->Equal(Ts[i]->GetChr(), Ts[i]->GetPos()) )
	  {
	    nid = i; 
	    break;
	  }

      if (nid == -1) break;
      if (Ts[nid]->Smaller(chr, pos)) //nid < (chr, pos)
	stntid = nid;
      else break;
    }
  if (nid == -1) return  GetTree(stntid, T); //no more data after stntid
  for (int i = stntid; i != nt; ++i)
    if (Ts[i]->Equal(chr, pos) == 1) {
      stntid = i;
      return GetTree(stntid, T);
    }

  int ntree = GetTree(stntid, T);  
  int ntree1 = GetTree(nid, &T[ntree]);  

  return ntree+ntree1;
}

//find the order of nodes in ARG tree and in likelihood samples
void CARG_m :: InitLikelihoodName(char ** lknames, int nslk)
{
  Match(alnnames, nsp, lknames, nslk, index);
}

int CARG_m::ReadNextTree_bgzip(int rchr, int rpos)
{
  if (stntid > 100) Remove(stntid-1); //remove the first trees              

  int pos = rpos - K100, chr = rchr;

  if (pos < 0) pos = 0;

  if (nt > 0)
    {
      pos = Ts[nt-1]->GetPos();
      chr = Ts[nt-1]->GetChr();
    }

  int stpos = pos;
  int ont = nt;

  buffer_t buf = {NULL,0,0};

  char tempc[1024];
  int tid, beg, end;

  while (nt - ont < K1 && pos -  stpos < M1)
    {
      sprintf(tempc, "%d:%d-%d", chr, pos+1, pos + K100);

      if ( ti_parse_region(bgzipf_id, tempc, &tid, &beg, &end) != 0)
	{
	  bgzf_close(bgzipf);
	  fprintf(stderr, "failed to parse the region.\n");
	  exit(1);
	}
      ti_fetch(bgzipf, bgzipf_id, tid,  beg, end,  &buf, fetch_func);

      for (int i = 0; i != buf.ntree; ++i) 
	{
	  //	  Ts[nt] = new CTree;
	  
	  NewTree();
	  
	  if (Ts[nt] == NULL)
	    {
	      printf("calloc error in ReadNextTree \n");
	      exit(1);
	    }

	  if (!Ts[nt]->Init(buf.data[i])) { delete Ts[nt]; return 1;}

	  free(buf.data[i]);

	  if (Ts[nt]->GetNumLeaf() != nleaf) {
	    printf("error: %d\t%d\n", Ts[nt]->GetChr(), Ts[nt]->GetPos());
	    printf("nleaf = %d ntreeleaf = %d\n", nleaf, Ts[nt]->GetNumLeaf());
	  }
	  nt ++;
	}

      buf.ntree = 0;
      pos += K100;
      //      free(buf.data);//free data
    }
  
  free(buf.data);
  return 0;
}

