#include "CCall.h"



CCall::CCall()
{
  blocksize = M1;
  inputtype = Q_BINARY;

  strcpy(argname_m, ""); strcpy(samples_m, "");
  strcpy(argname_f, ""); strcpy(samples_f, "");
  
  strcpy(fname, "");
  strcpy(fnout, "");
  chr = -1;

  onego = 0;
}


CCall::~CCall()
{
  for (int i = 0; i != nlks; ++i) free(likelihoodFileNames[i]);
}

void CCall :: Filter(char ** names, int ns)
{
  printf("filter: ns = %d\n", ns);
  int l = 0;
  for (int i = 0; i != nlks; ++i)
    {
      int ok = 0;
      for (int j = 0; j != ns; ++j) 
	if (strstr(likelihoodFileNames[i], names[j]) != NULL) ok = 1;

      if (ok)
	{
	  if (l < i)
	    {
	      free(likelihoodFileNames[l]);
	      likelihoodFileNames[l] = likelihoodFileNames[i];
	    }
	  l++;
	}
    }
  nlks = l;
}

int CCall :: NotMatch(char **s1, char** s2, int ns1, int ns2)
{
  if (ns1 != ns2) return 0;
  
  for (int i = 0; i != ns1; ++i)
    {
      int ok = 0;
      for (int j = 0; j != ns2; ++j)
	if (strcmp(s1[i], s2[i]) == 0) ok = 1;
      if (!ok) return 0;
    }
  return 1;
}


void CCall :: Call(void)
{
  printf("starting call SNP\n");



  //trees from Margarita ARGs
  CARG_m *argM = new CARG_m;//MARGARITA
  CARG_f* farg = new CARG_f; //CARGs_Fast ARG


  if (!argM->Init(argname_m, samples_m)) 
    {
      delete argM;
      argM = NULL;
    }

  if (!  farg->Init(argname_f, samples_f)) 
    {
      delete farg;
      farg = NULL;
    }
  
  if (argM != NULL && farg != NULL)
    {
      if (NotMatch(argM->GetNames(), farg->GetNames(), argM->GetNsample(), farg->GetNsample()))
	{
	  printf("samples in arg trees or fast arg tree do not meet: %s %s;\n", samples_m, samples_f);
	  exit(1);
	}
    }


  if (argM != NULL) Filter(argM->GetNames(), argM->GetNsample());
  if (farg != NULL) Filter(farg->GetNames(), farg->GetNsample());
  
  if (argM != NULL) argM->InitLikelihoodName(likelihoodFileNames, nlks);
  if (farg != NULL) farg->InitLikelihoodName(likelihoodFileNames, nlks);


  if (onego && argM == NULL && farg == NULL)
    {
      printf("-onego options must company with input ARGs, -ARG or -farg options\n");
      exit(1);
    }

  int i = st, nr = 0;
  while (1==1)
    {
      CDataLikelihood *d = new CDataLikelihood; //get a block data
      d->Init(likelihoodFileNames, nlks,  chr, i,  blocksize);

      //      if (argM != NULL) argM->ReadNext(d->GetChr(), d->GetPos(0));

      CARG_f* fargl = NULL; //loop arg = NULL
      CHaplotypes *H = NULL;//old haplotypes = NULL

	//no ARG trees, build Haplotypes from likelihood without linkage disequibilirum 
      if (argM == NULL && farg == NULL && fargl == NULL)  //no ARGs
	{
	  CFastARG *fastARG = new CFastARG;
	  CHaplotypes *H = d->Genotyping();
	  fargl = fastARG->Build(H);//build the first farg
	  fargl->InitLikelihoodName(likelihoodFileNames, nlks);//set names and number of samples for fargl
	  
	  delete H;
	  delete fastARG;
	}

      do {
	printf("\n\n\nStart a loop\n");
	CPhased *phase = new CPhased;
	//	CHaplotypes *H1 = phase->BuildHap(d, argM, farg, fargl);
	CHaplotypes *H1 = phase->BuildHap(d, argM, farg);
	delete phase;

	if (H == NULL) H = H1;
	else
	  {
	    int t = H1->Identical(H1);
	    delete H;
	    H = H1;
	    if (t == 1) break; //if two haplotypes are identical
	  }

	H->Display();
	if (onego) break;

	CFastARG *fastARG = new CFastARG;   
	CARG_f* fargl1=  fastARG->Build(H); //rebuild fargl
	delete fastARG;

	fargl1->InitLikelihoodName(likelihoodFileNames, nlks);
	delete fargl;
	fargl = fargl1;
	if (argM != NULL) argM->Reset();
      } while (1==1);


      if (fargl != NULL) delete fargl;  
      delete d;
      delete H;
      if (nr++ == 2) break;
    }

  delete argM;
  delete farg;
}




void CCall:: SetThreshold(double x) { threshold = x;}
//void CCall:: SetBlockSize(int x) { blocksize = x; }

//void CCall:: SetFileNames(char *fn) { strcpy(fname, fn);}
void CCall:: SetOutPut(char *fn) { strcpy(fnout, fn);}
//void CCall:: SetChrStEd(int rchr, int rst, int red) { st = rst; ed = red; chr = rchr;};


void CCall:: SetARG_M(char *i_argname, char *i_samples) 
{ 
  strcpy(argname_m, i_argname);
  strcpy(samples_m, i_samples);
}
  
void  CCall:: SetARG_f(char *i_argname, char *i_samples) 
{ 
  strcpy(argname_f, i_argname); 
  strcpy(samples_f, i_samples);
}





/*
void CCall :: PutLikelihoodNames(char *s)
{
  likelihoodFileNames[nlks] = (char*)calloc(strlen(s) + 10, sizeof(char));
  strcpy(likelihoodFileNames[nlks++], s);
}
*/

/*
void CCall ::ReadLikelihoodNames(char *fn)
{

  FILE *f = fopen(fn, "r");
  if (f == NULL)
    {
      printf("error when opening: %s\n", fn);
      exit(1);
    }

  char tempc[1024];
  nlks = 0;

  while (!feof(f))
    if (fscanf(f, "%s", tempc) == 1) 	PutLikelihoodNames(tempc);
  fclose(f);
  printf("Number of likelihood sample, nlks,  = %d\n", nlks);
}


*/
