#ifndef __CComputing__
#define  __CComputing__
#include "../Common/Const.h"

#ifndef MaxFlankingTree
#define MaxFlankingTree 100
#endif

class CComputing
{

private:
  int nsample;
  int rid;
  ProbOneSite *p;
  double maxlog;
  char *homs;

  double ** Conf;

  char **gen[K10];
  int  ngs[K10];


  //  char ** Gens[MaxFlankingTree];  
  //  double * ProbPP[MaxFlankingTree];

  double ** ProbGen[MaxFlankingTree]; //Prob [tk] [e] [genotype]
  double ** ZZ[MaxFlankingTree];  //ZZ[tk] [sample i] {[0][1][2]} 

  double ** PGhet;
  double ** PGhom;

  double ** pH;


  char *hap;
  char* conf;


  double LogProb(char *gen, int id1, int id2);
  //  void CalculatePP(char** gen, int ng, double *PriorP);

  void CalculatePP(char** gen, int ng, double *PriorP);
  void CalculateZ (char** gen, int ng, double *PP,  double **Z);

  void PostView(char *hap);

  //  void GenotypingComputation(double *phom, double *phet);

  void GenotypingComputation(double *phom, double *phet, char &al1, char &al2, char &conf);

  //  void CalculatePP(char** gen, int ng, double *PriorP);

  double ProbHom[NA];

  int nleaf;
  int nseq;
  int nt;
  double ptk;
  double UnSeenProb;


  double  LogProb(int id); //logprob hom id

  CTree ** T;

public:

  CComputing(int );//{ ProbGen = NULL; Conf = NULL;}
  ~CComputing();

  void Init(ProbOneSite *pp, int idref, CTree** T, int nt);


//double CComputing::ScanTree(CTree **T, int nt, int rid)
//Calculate the P(D|mut, T)
  double CalProbMutation(int rid);
  void ComputingProb(char** gen, int ng);
  

  //  double ProbMutation(CTree **T); // 
  //  double ProbMutation(void);

  double Prob(char *gen, int id1, int id2);
  double ProbNoMutation(void);

  void Genotyping(void);

  void VertPhased(double *p, char &al, char&conf);
  void PhasedGenotype(void);

  char * GetHap(void) { return hap;}
  char * GetConf(void) { return conf;}

};




CComputing::CComputing(int nsample_in)
{
  nsample = nsample_in;
  nleaf = 2*nsample;
  nseq =  4*nsample-2;
  

  double tempd;
  for (int i = 1; i != 2*nsample + 2; ++i) tempd += 1/(double)i;
  UnSeenProb = (1 /(double) (2*nsample + 1)) / tempd;

  //ProbGen[i (tree)][k (string)][x-->y]
  for (int i = 0; i != MaxFlankingTree; ++i)    ProbGen[i] = Calloc(4*nsample, NA*NA);  //ProbGen[i][e][g]: Prob of T_i, e, genotype
  for (int i = 0; i != MaxFlankingTree; ++i)    ZZ[i]      = Calloc(nsample, 3);  //ZX[i] [e] [0][1][2]

  Conf = Calloc(2*nsample, NA*NA);

  PGhet = Calloc(nsample, NG);
  PGhom = Calloc(nsample, NA *NA);
  pH = Calloc(2*nsample, NA * NA);

  hap = CallocChar(nleaf * 2);
  conf = CallocChar(nleaf *2);
}


CComputing::~CComputing()
{

  for (int i = 0; i != MaxFlankingTree; ++i) Free( ProbGen[i], 4*nsample);
  for (int i = 0; i != MaxFlankingTree; ++i) Free( ZZ[i], nsample);
  Free(Conf, 2*nsample);

  Free(PGhet, nsample);
  Free(PGhom, nsample);
  Free(pH, 2*nsample);

  free(hap);
  free(conf);
}




void CComputing::Init(ProbOneSite *pp, int idref, CTree** Tin, int i_nt) 
{
  p = pp;
  nt = i_nt;
  rid = idref;

  ptk = 1 / (double) nt;

  double maxlog = -(1E+30);

  char *maxp ;
  int mx, my;

  T = Tin;




  for (int i = 0 ; i != nt; ++i)
    {
      char **gens = T[i]->GetFig();
      nseq = T[i]->GetNumGens();

      double **probgen = ProbGen[i];

      //      printf("nseq = %d\n", nseq);

      for (int j = 0; j != nseq; ++j)
	for (int x = 0; x != NA; ++x)
	  for (int y = 0; y != NA; ++y)
	    if (x != y)
	      {

		//		printf("%d\t%s\n", j, gens[j]);

		probgen[j] [x*NA + y] = LogProb(gens[j], x, y);

		if ( probgen[j] [x*NA + y] > maxlog) {
		  maxlog =  probgen[j] [x*NA + y];
		  maxp = gens[j];
		  mx = x; 
		  my = y;
		}
	      }
    }
  //  printf("return here in Init\n");  

  for (int i = 0 ; i != nt; ++i)        
    {
      double **probgen = ProbGen[i];              
      for (int j = 0; j != nseq; ++j)    
	for (int x = 0; x != NA; ++x)  
	  for (int y = 0; y != NA; ++y)  
	    if (x != y)     
	      probgen[j][x*NA + y] = EXP(probgen[j][x*NA + y] - maxlog);
    }


  for (int i = 0; i != NA; ++i)
    ProbHom[i] = EXP(LogProb(i) - maxlog);
}



void CComputing :: Genotyping(void)
{
  for (int i = 0 ; i != nt; ++i)
    {
      char** gens = T[i]->GetFig();
      double** Z = ZZ[i];
      CalculateZ (gens, nseq, T[i]->GetPP(), Z);
    }

  //  printf("ok herer\n");

  for (int sid = 0; sid != nsample; ++sid)    
    {
      for (int i =0 ; i != NG; ++i)    PGhet[sid][i] = 0;
      for (int i = 0; i != NA*NA; ++i) PGhom[sid][i] = 0;
    }


  //p(D,gi=ab) = sum_k sum_e sum_{x,y} p(tk) p(e) p(D,gi=ab| e:x->y

  for (int k = 0; k != nt; ++k) //tree t_k
    {
      char** gens = T[k]->GetFig();
      double *PP = T[k]->GetPP();

      double** probgen = ProbGen[k];
      double** Z = ZZ[k];      

      for (int seqid = 0; seqid != nseq; ++seqid)
	for (int a = 0; a != NA; ++a)
	  for (int b = 0; b != a; ++b)
	    {
	      double t =   TsTv[a][b]* probgen[seqid][a*NA + b] 
		         + TsTv[b][a]* probgen[seqid][b*NA + a];
	      
	      t *=  PP[seqid];
	      t /= 2;

	      if (t < epsilonRM) continue; //too small, no need to consider
	      for (int sid = 0; sid != nsample; ++sid)
		if (gens[seqid][sid*2] != gens[seqid][sid*2 +1])    
		  PGhet[sid][UIB[a][b]] += t;
	    }

      //PGhom, aa = r
      for (int seqid = 0; seqid != nseq; ++seqid)
	{
	  for (int a = 0; a != NA; ++a)
	    {
	      double temp0 = 0, temp2 = 0;	      
	      for (int x = 0; x != NA; ++x) 
		if (x != a)
		{
		  temp0 += TsTv[a][x] * probgen[seqid] [a*NA +x];
		  temp2 += TsTv[x][a] * probgen[seqid] [x*NA +a];
		}

	      temp0 *=  PP[seqid];
	      temp2 *=  PP[seqid];

	      if (temp0 + temp2 < epsilonRM) continue;

	      for (int sid = 0; sid != nsample; ++sid)  
		if (gens[seqid][sid*2] == gens[seqid][sid *2 +1])  
		  {
		    if (gens[seqid][sid*2]  == '0') PGhom[sid][a*NA+a] +=  temp0;
		    if (gens[seqid][sid*2]  == '1') PGhom[sid][a*NA+a] +=  temp2; 
		  }
	    }

	}

      //PGhom, aa != r
      for (int seqid = 0; seqid != nseq; ++seqid)
	for (int a = 0; a != NA; ++a)
	  for (int r = 0; r != NA; ++r)
	    if (a != r) //aa != r
	      {
		double temp0 =  PP[seqid]  * TsTv[a][r] *   probgen[seqid][a*NA + r];
		double temp2 =  PP[seqid]  * TsTv[r][a] *   probgen[seqid][r*NA + a];

		if (temp0 + temp2 < epsilonRM) continue;

		
		for (int sid = 0; sid != nsample; ++sid)  
		  if (gens[seqid][sid*2] == gens[seqid][sid *2 +1])  
		    {
		      if (gens[seqid][sid*2]  == '0')  PGhom[sid][a*NA+r] += temp0;
		      if (gens[seqid][sid*2]  == '1')  PGhom[sid][a*NA+r] +=  temp2; 
		    }
	      }
    }



  for (int sid = 0; sid != nsample; ++sid)    
    for (int a = 0; a != NA; ++a)    
      for (int b = 0; b != a; ++b) PGhet[sid][UIB[a][b]] *= ptk;

  for (int sid = 0; sid != nsample; ++sid)  
    for (int a = 0; a != NA; ++a)
      for (int r = 0; r != NA; ++r)  PGhom[sid][a*NA +r] *= ptk;


  for (int sid = 0; sid != nsample; ++sid)  
    for (int a = 0; a != NA; ++a)
      for (int r = 0; r != NA; ++r)
	if (a != r)
	  PGhom[sid][a*NA +r] = UnSeenProb*ProbHom[a] + (1-UnSeenProb)* PGhom[sid][a*NA +r];

  for (int sid = 0; sid != nsample; ++sid)      
    GenotypingComputation(PGhom[sid], PGhet[sid], hap[2*sid], hap[2*sid+1], conf[sid]);


  //  printf("nsample = %d\n", nsample);

  hap[2*nsample] = '\0';
  conf[nsample] = '\0';
  PostView(hap);
  //  printf("%s\t%s\n\n", hap, conf);
}





void CComputing :: GenotypingComputation(double *phom, double *phet, char &al1, char &al2, char &conf )
{
  double probb;
  double gt[10], pG[10];

  for (int i =0; i != NG; ++i) pG[i] = 0;
  for (int r = 0; r != NA; ++r)
    {
      if (r != rid) probb = RefErr / 3;
      else probb = 1-RefErr;

      for (int a = 0; a != NA; ++a)	gt[a] =  phom[a * NA + r];  

      for (int a = 0; a != NA; ++a)
	for (int b = 0; b != a; ++b)
	  if (r != a && r != b) gt[ UIB[a][b]] = 0;
	  else gt[ UIB[a][b] ] = phet[ UIB[a][b] ]; 
      
      Norm(gt);
      //      printf("probb = %e\n", probb);
      //      DISPLAY(gt, NG);

      for (int i = 0; i != NG; ++i) pG[i] += gt[i] * probb;
    }
  Norm(pG);
  //  printf("\n");
  //  DISPLAY(pG, NG);

  int id = MaxID(pG);
  al1 = H1[id];
  al2 = H2[id];
  conf = Prob2Phred(1-pG[id]);


}


void CComputing :: PhasedGenotype(void)
{

  for (int i =0; i != 2*nsample; ++i)
    for (int j = 0; j != NA*NA; ++j) pH[i][j]  = 0;
  
  for (int k = 0; k != nt; ++k)
    {
      //      char** gens = Gens[k];
      char** gens = T[k]->GetFig();

      //double*  PP = ProbPP[k];
      double*  PP = T[k]->GetPP();

      double** probgen = ProbGen[k];


      //P(d,g=aa, a == r	  
      for (int e = 0;  e != nseq; ++e)
	for (int a = 0; a != NA; ++a)
	  {
	    double tempR = 0, tempnotR = 0;
	    for (int x = 0; x != NA; ++x)
	      if (x != a) 
		{
		  tempR += TsTv[x][a] * probgen[e][x*NA + a];
		  tempnotR += TsTv[a][x] * probgen[e][a*NA + x];
		}
	    tempR *=  PP[e];
	    tempnotR *= PP[e];
	    if (tempR + tempnotR < epsilonRM) continue;

	    for (int i = 0;  i != 2*nsample; ++i)
	      if (gens[e][i] == '1') pH[i][a*NA + a] +=  tempR;
	      else pH[i][a*NA + a] += tempnotR;
	  }

      //P(d,g=aa, a != r
      for (int e = 0; e != nseq; ++e)
	for (int a = 0; a != NA; ++a)
	  for (int r = 0; r != NA; ++r)
	    if (r != a) 
	      {
		double tempR    = PP[e]*TsTv[r][a]*probgen[e][r*NA +a];
		double tempnotR = PP[e]*TsTv[a][r]*probgen[e][a*NA +r];

		if (tempR + tempnotR < epsilonRM) continue;

		for (int i = 0;  i != 2*nsample; ++i)
		  if (gens[e][i] == '1') pH[i][r*NA + a] += tempR;
		  else pH[i][r*NA + a] += tempnotR;
	      }



    }

  for (int i = 0; i != 2*nsample; ++i)
    for (int a = 0; a != NA; ++a)
      for (int r = 0; r != NA; ++r) 
	pH[i][r*NA+a] *= ptk;

  for (int i = 0; i != 2*nsample; ++i)
    for (int a = 0; a != NA; ++a)
      for (int r = 0; r != NA; ++r) 
	if (a != r)
	  pH[i][r*NA+a] = UnSeenProb * ProbHom[a]  + (1-UnSeenProb) *  pH[i][r*NA+a];


  for (int i = 0; i != 2*nsample; ++i)
    VertPhased(pH[i], hap[i], conf[i]);
  PostView(hap);
  View(conf, nsample);
}



void CComputing::PostView(char *hap)
{
  for (int sid = 0; sid != nsample; ++sid)
    if (p->data[sid] == 0)
      {
	hap[2*sid] += 32;
	hap[2*sid+1] += 32;
      }
  ViewRef(hap, nsample, rid);
  View(hap, nsample);
}


void CComputing::VertPhased(double *p, char&al, char &conf)
{

  //  Display(p, NA*NA);

  double probb;
  double probH[NA];

  for (int a = 0; a != NA; ++a) probH[a] = 0;

  for (int r = 0; r != NA; ++r)
    {
      double *pg = &p[r*NA];
      Norm(pg, NA);

      if (r != rid) probb = RefErr / 3;
      else probb = 1-RefErr;
      for (int a = 0; a != NA; ++a) probH[a] += probb * pg[a];
    }

  Norm(probH, NA);

  int id = MaxID(probH, NA);
  
  al = DNA[id];
  conf = Prob2Phred(1 - probH[id]);

}




double CComputing :: LogProb(char *gen, int id1, int id2)  
{
  int hetid = GenID(id1, id2);
  double temp = 0;
  //  printf("hetid = %d, nsample = %d\n", hetid, nsample);
  for (int i = 0; i != nsample; ++i)
    if (p->data[i])
      {
	int id = hetid;

	if (gen[i*2] == gen[i*2+1])
	  if (gen[i*2] == '0') id = id1;
	  else id = id2;

	temp += p->log_p_D[i*NG + id];
      }
  return temp;
}



double CComputing :: LogProb(int id) //logprob hom id
{
  double temp = 0;

  for (int i = 0; i != nsample; ++i)
    if (p->data[i])
      temp += p->log_p_D[i*NG + id];
  return temp;
}






double CComputing::ProbNoMutation(void)
{
  double out = 0;
  for (int r = 0; r != NA; ++r)
    if (r == rid) out += ProbHom[r] * (1-RefErr);
    else out += ProbHom[r] * (RefErr)/3;
  return out;
}


//double CComputing::ScanTree(CTree **T, int nt, int rid)
//Calculate the P(D|mut, T)
double CComputing::CalProbMutation(int rid)
{
  double out = 0;   //  out = 1/(2m+1) * sum_r p(r) { sum_{a<>r} TsTv[a][r] p(D|h=a) }

  for (int r = 0; r != NA; ++r)
    {
      double probr = RefErrA[rid][r];
      double temp = ProbHom[r];//temp = Prob(D|h=r) + sum_{a <> r} p(D|h=a)

      for (int a = 0; a != NA; ++a)
	if (a != r) temp += TsTv[a][r]  * ProbHom[a];
      
      out += temp * probr;
    }

  out *=  UnSeenProb;

  double out2 = 0; //out2 = 1/2 * (1-1/(2m+1)) \sum_r p(r) \sum_t_k p(t_k) \sum e p(e) \sum_{a <> r} p(a,r) [ p(D|e(a,r) t_k)+p(D|e(r,a), t_k)

  //out2 = sum_r p(r) p(D|r, T)
  //  double ptk = 1 / (double) nt;
  
  for (int r = 0; r != NA; ++r)
    {
      double probr = RefErrA[rid][r];
      double temp3 = 0; //p(D|r,T) = sum_{tk} p(tk) p(D|mut, r, t_k)

      for (int k = 0; k != nt; ++k)
	{
	  //	  char ** gen = Gens[k];  
	  char ** gen = T[k]->GetFig();

	  //	  double *pp =  ProbPP[k];
	  double *pp = T[k]->GetPP();

	  double **probgen = ProbGen[k];

	  double temp2 = 0; //p(D|r,tk) = sum_e p(e)p(D|e,tk, r)
	  for (int e = 0; e != nseq; ++e)
	    {
	      double temp1 = 0;  //p(D|e,tk, r) = sum_a p(a,r) [ p(D|e:a->r) + p(D|e:r->a]
	      for (int a = 0; a != NA; ++a)
		temp1 +=  TsTv[a][r] *  ( probgen[e][r *NA + a] +  probgen[e][a *NA + r]);

	      temp2 += temp1 * pp[e];
	    }
	  temp3 += temp2 * ptk;
	}

      out2 +=  RefErrA[rid][r] * temp3;
    }

  out2 *= .5;
  out2 *= (double) (1 - 1/(double) (2*nsample + 1));

  return out + out2;
}

/*
void CComputing::CompConf(void)
{
  for (int i =0; i != nt; ++i)
    {
      char ** gen = Gens[i];//t->SetMuts(1, ng);
      double *pp =  ProbPP[i];
      double **probgen = ProbGen[i];
      AddConf(gen, nseq, pp, ProbGen[i]);
    }
}

void CComputing::AddConf(char **gen, int ng, double *pp, double ** ProbGen)
{
  double Prr =(double) 1/3;
  double Par = 1;

  int a, r;

  for (int id =0; id != ng; ++id)
    {
      for (int x = 0; x != NA; ++x)
	for (int y = 0; y != NA; ++y)
	  if (x != y)
	    {
	      double p = ProbGen[id][x*NA+y] * pp[id];

	      //	      printf("%e\n",  ProbGen[id][x*NA+y]);
	      if (ProbGen[id][x*NA+y] < 1e-10) continue;

	      //	      printf("%s\n%d\t%d\n", gen[id],x, y);

	      //a == r
	      for (int i = 0; i != 2 * nsample; ++i)
		{
		  if (gen[id][i] == '0') r = x;
		  else r = y;

		  Conf[i][r*NA + r] += Prr * p;
		}

	      // a != r
	      for (int i = 0; i != 2*nsample; ++i)
		{
		  if (gen[id][i] == '0') {a = x; r = y;}
		  else {a = y; r = x;}
		  Conf[i][a*NA + r] += Par * p;
		}
	    }
    }
}



void CComputing::ConvertHaplotype(char*s1, char* s2)
{
  return;
  double const refcor = 1 -  RefErr;
  double const referr = RefErr/3;

  double** p = Calloc(2*nsample, NA);
  
  //  printf("rid = %d\n", rid);
  for (int i = 0; i != 2*nsample; ++i)
    {
      double *conf = Conf[i];

      for (int r = 0; r != NA; ++r)
	{
	  double sum = 0;
	  for (int a = 0; a != NA; ++a)	    sum += conf[a*NA + r];

	  if (sum > 0)
	    for (int a = 0; a != NA; ++a)   conf[a*NA + r] /= sum;
	}

      for (int a = 0; a != NA; ++a)
	for (int r = 0; r != NA; ++r)
	  if (r == rid) p[i][a] += conf[a*NA + r] * refcor;
	  else p[i][a] += conf[a*NA + r] * referr;
    }


  for (int i = 0; i != 2*nsample; ++i)
    {
      int ma = 0;
      Norm(p[i], NA);

      for (int a = 0; a != NA; ++a)
	if (p[i][a] > p[i][ma]) ma = a;


      //      for (int j =0; j != NA; ++j) printf("%e\t", p[i][j]);
      //      printf("%e\n", p[i][ma]);

      if (p[i][ma] < .25)
	{
	  printf("error: %d %d %e\n", i, ma, p[i][ma]);
	  exit(1);
	}
      s1[i] = DNA[ma];
      s2[i] = Prob2Phred(1-p[i][ma]);
    }

  Free(p, 2*nsample);
  s1[2*nsample] = '\0';
  s2[2*nsample] = '\0';
}
*/


void CComputing::CalculateZ(char** gen, int ng,  double *PP, double **Z)
{
  return;
  for (int sid = 0; sid != nsample; ++sid )  //sample id
    for (int i = 0; i != 3; ++i) Z[sid] [i] = 0;

  for (int seqid = 0; seqid != ng; ++seqid) //sequence id
    for (int sid = 0; sid != nsample; ++sid )  //sample id
      {
	int zid = (gen[seqid] [sid*2] - '0') + (gen[seqid] [sid*2 + 1] - '0');
	Z[sid][zid] += PP[seqid];
      }
}


void CComputing::CalculatePP(char** gen, int ng, double *PriorP)
{
  //  double *PriorP = (double*)calloc(ng+10, sizeof(double));

  double pp[K10]; //proiorProb[n] = 
  int    Ind[K10]; //ind[i]: number of allele of gen[i]
  int    count[K10]; //count[i]: how many gens with i alleles

  for (int i = 0; i != 2*nsample;  ++i) count[i] = 0;

  for (int i = 0; i != ng;  ++i)
    {
      int nc = 0;
      for (long j = 0; j != 2*nsample; ++j)
	if (gen[i][j] == gen[i][0]) nc ++;

      if (nc > nsample) nc = 2*nsample - nc;

      count[nc]++;
      Ind[i] = nc;
    }


  for (long i = 1; i != 2*nsample; ++i)
    if (count[i] > 0) pp[i] = 1/(double) i + 1/(double) (2*nsample -i);
    else  pp[i]  = 0;
  pp[0] = 0;
  
  double sum = 0;
  for (long i = 1; i != 2*nsample; ++i) sum += pp[i];
  for (long i = 1; i != 2*nsample; ++i) pp[i] /= sum;
	     
  for (int i = 0; i != 2 *nsample; ++i) 
    if (count[i] > 0) pp[i] /= (double) count[i];


  pp[0] = 0;
  pp[2*nsample] = 0;

  for (int i =0; i != ng; ++i)    PriorP[i] = pp [ Ind[i] ];
}



/*
void CComputing :: Display(void)
{
  for (int i = 0; i != nsample; ++i)
    {
      for (int id = 0; id != 10; ++id) printf("%f\t", p->log_p_D[i*NG+id]);
      printf("\n");
    }
}
*/

#endif

