#include "CDataLikelihood.h"




CDataLikelihood :: ~CDataLikelihood()
{
  if (data != NULL) free(data);
  if (pos != NULL) free(pos);
  if (bfprob != NULL) free(bfprob);
  free(onelines);
}



CDataLikelihood :: CDataLikelihood()
{
  data = NULL;
  ns = 0;
  nm = 0; 
  onelines = (COneLine*)calloc(K10, sizeof(COneLine));
  pos = NULL;

  double log10 = log(10);

  for (int i = 0; i != 1000; ++i)
    {
      double x = i;
      Phred2Prob[i] = exp((-x/10) * log10);
      Phred2LogProb[i] = -x/10 * log10;
    }
  bfprob = NULL;
}

void CDataLikelihood :: Init(char **fname, int rns,  int rchr, int rst, int rbl)
{
  ns = rns;
  chr =  rchr;
  st = rst;
  bl = rbl;

  samplenames = fname;
  if (ns == 0)
    {
      printf("No sample likelihood. Using -k <file name> or -n <list of samples name>\n");
      exit(1);
    }
  
  for (int i =0; i != ns; ++i) onelines[i].Init(samplenames[i], Q_BINARY);
  
  if (chr != -1)
    for (int i =0; i != ns; ++i)
      onelines[i].Seek(chr, st);

  int mchr, mpos;
  nm = 0;
  int mnm = 10000;

  if (data != NULL) free(data);
  data = (unsigned char*)malloc(mnm * ns * NG);

  if (data == NULL)
    {
      printf("Malloc memory fail\n");
      exit(1);
    }

  if (pos != NULL) free (pos);
  pos = (int*)malloc(mnm * sizeof(int));

  if (pos == NULL)
    {
      printf("Malloc memory false\n");
      exit(1);
    }

  do {

    if (nm + 10 >= mnm)
      {
	mnm += 50000;
	data = (unsigned char*)realloc(data, mnm * ns * NG * sizeof(char));
	pos = (int*)realloc(pos, sizeof(int) * mnm);
      }

    mchr = 100;
    for (int i =0; i != ns; ++i)
      if (onelines[i].Compare(mchr, mpos) == 1) //(mchr, pos) > online[i] <chr, pos>
	onelines[i].GetChrPos(mchr, mpos);

    if (mchr == -1 || mchr == 100) break; //end of data

    unsigned char *p = &data[nm*ns*NG];
    for (int i =0; i != ns * NG; ++i) p[i] = 0;

    for (int i = 0; i != ns; ++i)
      if (onelines[i].Compare(mchr, mpos) == 0) //find the 
	{
	  onelines[i].GetLK(& p[i*NG]);
	  onelines[i].Read();

	  //norm, p[i] = 0 for the maximum
	  int t = 10000;
	  for (int k =0; k != NG; ++k)
	    if (p[i*NG + k] < t) t = p[i*NG + k];

	  for (int k =0; k != NG; ++k)	     p[i*NG + k] -= t;
	}

    if (chr != -1)
      if (mchr < chr || (mchr == chr && mpos < st)) continue; 

    pos[nm++] = mpos;
    if (chr == -1) chr = mchr;
  } while ( (mchr == rchr || rchr == -1) && mpos  - pos[0] < bl);

  


  printf("Finish reading data: ns = %d\tnmarker = %d\n",ns,  nm);

}





void CDataLikelihood :: Display(void)
{
  for (int i = 0; i != nm; ++i)
    {
      printf("\n\n%d\t", pos[i]);
      for (int j =0; j != ns; ++j)
	{
	  unsigned char *p = &data[i*NG*ns + j*NG];
	  for (int k = 0;  k != NG; ++k) printf("\t%d", (int) p[k]);
	}
    }
  printf("\n\n");
}




CHaplotypes* CDataLikelihood :: Genotyping(void)
{
  CHaplotypes *H = new CHaplotypes;
  
  if (H == NULL) 
    {
      printf("error new Haplotypes\n");
      exit(1);
    }
  printf("ns = %d, nm = %d\n", ns, nm);
  H->Init(ns, nm, samplenames); //Init

  char DNA[] = "ACGT";

  char *CharArr[5];
  for (int i = 0; i != 5; ++i) CharArr[i] = NULL;
  
  CharArr[0] = (char*)calloc(2*ns+1, sizeof(char));
  CharArr[0][2*ns] = '\0';

  for (int k = 0; k != nm; ++k)
    {
      COneSite *s = new COneSite;
      char *haps = CharArr[0];
      for (int i = 0; i != 2*ns; ++i)	haps[i]  = '?';  //no data

      double confs[NG];

      for (int i = 0; i != ns; ++i)
	{
	  unsigned char *p = &data[k*ns*NG + i * NG];
	  //check if data exist
	  int ok = 0;
	  for (int j = 0; j != NG; ++j)
	    if (p[j] == 0) ok = 1;
	  if (!ok) continue; //no data

	  for (int j = 0; j != NG; ++j) confs[j] = Phred2Prob[p[j]];
	  
	  int l = 0;
	  for (int k1 = 0; k1 != NA; ++k1)
	    for (int k2 = k1; k2 != NA; ++k2)
	      if (k1 == k2) confs[l++] *= .99;
	      else confs[l++] *= .01;
	  //	  Norm(confs);

	  double mprob = 0;
	  
	  l = 0;
	  for (int k1 = 0; k1 != NA; ++k1)
	    for (int k2 = k1; k2 != NA; ++k2)
	      {
		if (mprob < confs[l]) {
		  haps[i*2]   = DNA[k1];
		  haps[i*2+1] = DNA[k2];
		  mprob = confs[l];
		}
		l++;
	      }
	}
      s->SetCharArr(CharArr, 5);
      s->SetChrPos(chr, pos[k]);
      H->SetSite(k, s, 0);
    }
  return H;
}


double * CDataLikelihood :: GetDataFrom(int k, double *d)
{
  if (bfprob == NULL) bfprob = (double*)calloc(ns*NG, sizeof(double));
  unsigned char *p = &data[k*ns*NG];
  for (int i = 0; i != ns * NG; ++i) bfprob[i] = d[ (unsigned char) p[i]];
  return bfprob;
}


