#include "CFastARG.h"




//initialize
CFastARG::CFastARG()
{
  nmarker  = 0;
  nh = 0;
  marker = NULL;
  count = NULL;
  pt = (OnePoint*)calloc(K1000, sizeof(OnePoint)); //maximum 1.000.000 nodes
  BuiltARG = 0;
}

//free parameters
CFastARG::~CFastARG()
{
  if (marker != NULL) free(marker);

  for (int i = 0; i != n; ++i)
    if (nodes[i] != NULL) delete nodes[i];

  free(pt);
}




//read data from file

/*
<n: number of haplotypes> <m: number of marker>
<chr> <pos> [n values]

100     12301
20      43000108        0100100011110011000011000000000110110000110000101111101100011000101000110111111011110010001000110001
20      43000178        0010101001100011000011111000001110110011000011101111111000011100000100111011110011100000000100001110
20      43000202        1111111111111111111111111111101111111111111111111111111111111111111111111111111111101111111101101111
20      43000209        0010101001100011000111111000001110110011000011011110101100011000000000111011110011100100000100000110
20      43000282        0000000000000000000000000000010000000000000000000000000000000000000000000000000000000000000011000000
20      43000465        0000000000000000000000000000000000000000000000000000010010000000000000000100001000000000100000000000
*/




CARG_f* CFastARG :: Build(CHaplotypes* Hin)
{
  printf("Starting Building ARG from halotype data\n");

  CHaplotypes* H = Hin->Extract4ARG();
  printf("finish get H\n");

  char ** haps = H->GetHaps();
  int *pos = H->GetPos();

  printf("nsp = %d, number of marker = %d\n", H->GetNsample(),  H->GetNmarker());
  Init(H->GetNsample() * 2, H->GetNmarker(), haps, pos, H->GetChr());
  
  CARG_f* arg = Export2CARG_f(H);
   //  arg->Display();

   delete H;
   Free(haps, H->GetNsample() * 2);
   free(pos);

   printf("Building farg tree: Done\n");
   return arg;
}
void CFastARG::Init(char *fn)
{
}

/*
void CFastARG::Init(char *fn)
{
   FILE *f = fopen(fn, "r");
   if (f == NULL) 
     {
       printf("error: %s\n", fn);
       exit(1);
     }


   char tempc[K10];
   int n, m;
   fscanf(f, "%d %d", &n, &m); //read n and m

   //  m = 1000;

   char **S = (char**)calloc(n, sizeof(char*)); //initialize S
   printf("%d\t%d\n", n, m);

   for (int i = 0; i != n; ++i) S[i] = (char*)calloc(m + 10, sizeof(char));
   int *rmark = (int*)calloc(m, sizeof(int));

   int chr;

   for (int l = 0; l != m; ++l)
     {
       int tempchr;
       if (fscanf(f, "%d %d", &tempchr, &rmark[l]) != 2) break;
       fscanf(f, "%s", tempc);
       for (int i =0; i != n; ++i) S[i][l] = tempc[i];
       chr = tempchr;
     }

   for (int i = 0; i != n; ++i) S[i][m] = '\0';
   Init(n, m, S, rmark, chr);
   for (int i = 0; i != n; ++i) free(S[i]);
   free(S);
   free(rmark);
   fclose(f);
   }


*/





/*
n: number of haplotypes
m: number of markers

S: haplotypes[n][m]: 
S[i][j] = '0', '1', '2' (unphased) or '3' (missing) where i: haplotypes, j: marker
M: markers: m markers
ichr: chrsomosome
*/

void CFastARG::Init(int in, int m, char **S, int *M,  int ichr)
{
  BuiltARG = 0;
  chr = ichr; //get chr
  //get marker
  if (marker != NULL) free(marker);

  nmarker = m;
  marker  = (int*)calloc(m, sizeof(int));
  for (int i = 0; i != m; ++i)    marker[i] = M[i];


  //free nodes
  for (int i = 0; i != nh; ++i)
    if (nodes[i] != NULL) delete nodes[i];

  //initialize number of haplotypes, nnodes, nactive nodes
  nh = in;
  nn = in;
  n  = in;

  nactivenodes = in;
  for (int i = 0; i != n; ++i) activenodes[i] = i;

  for (int i = 0; i != n; ++i)
    {
      nodes[i] = new CNode;
      S[i][m] = '\0';
      nodes[i]->Init(i, 0, nmarker-1, S[i]);
    }

  if (count != NULL) free(count);
  count = (int*)calloc(nmarker*4, sizeof(int));

  for (int i = 0; i != nmarker * 4; ++i) count[i] = 0;
  for (int i = 0; i != nh; ++i) ReCount(nodes[i]->GetHap(), 1);
  for (int i = 0; i != n; ++i)  pt[i].st = 0;
  for (int i = 0; i != n; ++i)  pt[i].ed = nmarker-1;
  for (int i = 0; i != n; ++i)  Initpt(i);
}



//find the nearest neighbor of id, dist
void CFastARG::Initpt(int id)
{
  char *s1 = nodes[id]->GetHap();

  pt[id].dist = -1;
  pt[id].nei = -1;

  for (int ii = 0; ii != nactivenodes; ++ii)
    {
      int i = activenodes[ii];
      if (i == id) continue;
      char *s2 = nodes[i]->GetHap();
      int l = CountLeft(s1, s2, pt[id].st, pt[id].ed, pt[i].st, pt[i].ed);

      if (l > pt[id].dist)
	{
	  pt[id].dist = l; //dist
	  pt[id].nei = i;  //neighbor
	}
    }
}

//recount count vector with haplotype s (c = -1: remove, c=1 add)
void CFastARG :: ReCount(char *s, int c)
{
  for (int i = 0; i != nmarker; ++i)
    count[i*4+ s[i] - '0' ] += c;
}


//remove node id.
//if rmhap = 1: remove id out of active list

void CFastARG :: Remove(int id, int rmhap)
{
  if (rmhap) //remove nodes 
    {
      for (int i = 0; i < nactivenodes-1; ++i)
	if (activenodes[i] == id) activenodes[i] = activenodes[nactivenodes-1];
      nactivenodes --;
      nodes[id]->FreeHap(); 
      pt[id].dist = -1;
      pt[id].nei = -1;
    }
  
  //recount neighbor for nodes whose neighbor = id (because we just removed id)
  for (int ii = 0; ii != nactivenodes; ++ii)
    {
      int i = activenodes[ii];
      if (pt[i].nei == id) Initpt(i);
    }
}



//add new node id.
//if add = 1: add id out of active list. Otherwise, id is already in active list
void CFastARG :: Add(int id, int add)
{
  if (add) activenodes[nactivenodes ++]  = id;

//recount neighbor and distance for all nodes
  for (int i = 0; i != nactivenodes; ++i)    
    if (activenodes[i] != id) ReInit(activenodes[i], id);
  Initpt(id);
}




//recalculate (neighbor and dist) for node id when new node added, nid
void CFastARG :: ReInit(int id, int nid)
{
  char *s1 = nodes[id]->GetHap();
  char *s2 = nodes[nid]->GetHap();
  int l = CountLeft(s1, s2, pt[id].st, pt[id].ed, pt[nid].st, pt[nid].ed);
  
  if (l > pt[id].dist)
    {
      pt[id].dist = l; //start
      pt[id].nei = nid;//neighbor
    }
}


//count common part between s1 and s2 from the left
int CFastARG :: CountLeft(char *s1, char *s2)
{
  return CountLeft(s1, s2, 0, nmarker, 0, nmarker);
}


//count common part between s1 and s2 from the left
//where s1 = ['3'...'3' [st1 ... ed1] '3'..........'3']
//where s2 = ['3'.........'3' [st2 ... ed2] '3'....'3']
int CFastARG :: CountLeft(char *s1, char *s2, int st1, int ed1, int st2, int ed2)
{
  //  int st = st1;
  //  if (st < st2) st = st2;
  int st = MAX(st1, st2);
  int ed = MIN(ed1, ed2);

  //  int ed = ed1;
  //  if (ed > ed2) ed = ed2;

  ed ++; 
  if (st >= ed) return 0;

  int l = st;
  while ((l < ed) && (s1[l] + s2[l] != '0' + '1')) l++;

  
  for (int i = st; i != l; ++i) 
    if (s1[i] <= '2' && s2[i] <= '2')//unphased or phased haplotypes
      {
	if (l == ed) return nmarker;
	else return l;
      }
  return 0;
}

//find start and end point from *s of node id
void CFastARG :: InitSTED(int id, char *s)
{
  int l = 0;
  while (s[l] == '3' && l < nmarker) l++;
  pt[id].st = l;

  l = nmarker - 1;
  while (l > 0 && s[l] == '3') l--;
  pt[id].ed = l;
}

/*
//coalescence id1 and id2
//
//s:[min(st1, st2) - max(ed1, ed2)]
//            /       \
//           /         \
//          /           \
//    s1:[st - ed]   s2:[st-ed]
*/

void CFastARG::Coalescence(int id1, int id2)
{

  //if either id1 and id2 are not in the active list --> error
  if (!nodes[id1]->GetActive() ||  !nodes[id2]->GetActive())
    {
      printf("error (active in coalescence): %d\t%d\n", id1, id2);
      exit(1);
    }



  //new node
  CNode *p = new CNode;
  char *s = (char*)calloc(nmarker+10, sizeof(char));

  char *s1 = nodes[id1]->GetHap();
  char *s2 = nodes[id2]->GetHap();


  //if cannot Coalescence 
  if (!Equal(s1, s2))
    {
      printf("error in coalescence:id1, id2 %d %d\n%s\n%s\n", id1, id2, s1, s2);
      exit(1);
    }


  for (int i = 0; i != nmarker; ++i)
    if (s1[i] == '2' || s2[i] == '2')
    {
      int v = s1[i];
      if (s2[i] <= '1') v = s2[i];

      if (v > '1') v = '0'; //both missing or unphased

      if (s1[i] == '2') PutHap(id1, i, v);
      if (s2[i] == '2') PutHap(id2, i, v);
    }


  //build s: s[i] = s1[i] if (s1[i] != '3'; s[i] = s2[i] otherwise.
  for (int i = 0; i != nmarker; ++i)
    if (s1[i] != '3') s[i] = s1[i];
    else s[i] = s2[i];

  s[nmarker] = '\0';

  //note: st and ed here mean the location of id would be consider
  //id1: [st1 - ed1]
  //id2: [st2 - ed2] 
  
  int st =  MIN(nodes[id1]->GetST(), nodes[id2]->GetST()); 
  int ed =  MAX(nodes[id1]->GetED(), nodes[id2]->GetED()); 

  //  printf("coale\n%s\n%s\n%s\n\n", s1, s2, s);
  //  View(s1);
  //  View(s2);
  //  View(s);


  int newnode = n; //id of new node . 
  
  //put data into new node
  p->Init(newnode, st, ed, s, id1, id2, COAL);


  //initialize st and ed of s
  InitSTED(newnode, s);


  if (DISP) {
    printf("Coalescence: %d\t%d\n", id1, id2);
    nodes[id1]->Display();
    nodes[id2]->Display();
    p->Display();
  }

  //set tree, id1 and id2 have one common parent
  nodes[id1]->SetParent(1);
  nodes[id2]->SetParent(1);

  //put new node, p, into nodes
  nodes[n++] = p;

  //remove s1, s2
  ReCount(s1, -1);
  ReCount(s2, -1);

  //add s
  ReCount(s, 1);
  free(s);

  //remove id1, id2
  Remove(id1, 1);
  Remove(id2, 1);
  //add newnode
  Add(newnode, 1);
}

//check to string equal or not
//1: s1[i] == s2[i], exist s1[i] != '3' && s2[i] != '3'
int CFastARG :: Equal(char *s1, char *s2)
{
  for (int i = 0; i != nmarker; ++i)
    {
      if (s1[i] == '3' || s2[i] == '3') continue; //missing data
      if (s1[i] == '2' || s2[i] == '2') continue; //unphased data
      if (s1[i] != s2[i]) return 0; //phased data but mismatch
    }

  for (int i = 0; i != nmarker; ++i)
    if (s1[i] <= '1' && s2[i] <= '1') return 1;
    
  return 0;
}


void CFastARG :: PutHap(int id, int k, int v)
{
  int id1;
  if (id % 2 == 0) id1 = id + 1; //(0,1), (2,3), (4,5), .., (2n, 2n+1)
  else id1 = id - 1;

  int v1 = '0';
  if (v == '0') v1 = '1';


  if ( nodes[id]->GetAl(k) == 2)
    {
      nodes[id ]->ChangeHap(k, v);  
      count[k*4+2] --;
      count[k*4+ v - '0'] ++;
    }

  if (nodes[id1]->GetAl(k) == 2)
    {
      nodes[id1]->ChangeHap(k, v1);
      count[k*4+2] --;
      count[k*4+ v1 - '0'] ++;
    }
}

//coalescence for chr X
//coalescence (i and i+1)
void CFastARG::Coalescence_ChrX(void)
{
  for (int ii = 0; ii != nactivenodes; ++ii)
    {
      int i = activenodes[ii];
      if (Equal(nodes[i]->GetHap(), nodes[i+1]->GetHap())) Coalescence(i, i+1);
    }
}


//mutation
void CFastARG::Mutation(void)
{
  while (1==1)
    {
      //make all possible mutation
      for (int l = 0; l != nmarker; ++l)
	if ((count[l*4] == 1 || count[l*4+1] == 1) && (count[l*4+1] + count[l*4] > 1)  && count[l*4+2] == 0) 
	  Mutation(l, count[l*4], count[l*4+1]);

      //collect coalescence possibles
      que.Reset();
      for (int ii = 0; ii != nactivenodes; ++ii)
	{
	  int i = activenodes[ii];
	  if (pt[i].dist == nmarker) que.Push(i);
	}

      //      take a random
      int lci = que.GetRandom();
      if (lci == -1) break;
      Coalescence(lci, pt[lci].nei);
    }
}

//make mutation at l with n0 values 0 and n1 values 1
int CFastARG::Mutation(int l, int n0, int n1)
{
  int c = 1;
  if (n0 == 1)  c = 0;

  int nc = 0,  li = -1;
  for (int ii = 0; ii != nactivenodes; ++ii)
    {
      int i =  activenodes[ii];
      if (nodes[i]->GetAl(l) == c)  { nc ++; li = i;}
    }

  if (nc != 1) 
    {
      printf("error in mutation:%d\n", l);
      exit(1);
    }

  
  count[l*4+c] --;
  count[l*4+(1-c)] ++;
  nodes[li]->Mutation(l);

  Remove(li, 0);
  Add(li, 0);

  return li;
}


//recombination node id at x and x + 1
//
// s1:[st-x]      s2:[x+1:ed]
//       \          /
//        s:[st-ed]
void CFastARG::RecombinationPoint(int id, int x)
{
  if (DISP) printf("\n\nRecombination: %d\t%d\n", id, x);

  char *s = nodes[id]->GetHap();
  int st  = nodes[id]->GetST();
  int ed  = nodes[id]->GetED();

  char *s1 = (char*)calloc(nmarker+10, sizeof(char));
  char *s2 = (char*)calloc(nmarker+10, sizeof(char));


  for (int k = 0; k != nmarker; ++k) //if '2' --> '0'
    if (s[k] == '2') PutHap(id, k, '0');

  for (int i = 0; i != x; ++i) 
    {
      s1[i] = s[i];
      s2[i] = '3';
    }

  for (int i = x; i != nmarker; ++i) 
    {
      s1[i] = '3';
      s2[i] = s[i];
    }

  s1[nmarker] = '\0';
  s2[nmarker] = '\0';

  int nid = n;

  CNode *p1 = new CNode;
  p1->Init(n, st, x-1, s1,  id, -1, RECOMB);  //n-2

  nodes[n++] = p1;
  CNode *p2 = new CNode;
  p2->Init(n, x, ed, s2, id, -1, RECOMB);  //n-1
  nodes[n++] = p2;

  InitSTED(nid, s1);
  InitSTED(nid+1, s2);

  nodes[id]->SetParent(2);

  if (DISP)
    {
      nodes[id]->Display();
      p1->Display();
      p2->Display();
    }

  /*
  printf("\nRecombination\n");
  View(s);
  printf("---\n");
  View(s1);
  View(s2);
  */


  free(s1);
  free(s2);
  Remove(id, 1);
  Add(nid, 1);
  Add(nid+1, 1);
}



//recombination between id1 and id2 at x and y
void CFastARG::RecombinationPoint(int id1, int id2, int x, int y)
{

  //check if (id1 and id2) are active nodes.
  if (!nodes[id1]->GetActive() || !nodes[id2]->GetActive())
    {
      printf("error: %d %d in combination\n", id1, id2);
      exit(1);
    }
  //take a random node
  if (random() %2 == 0) { RecombinationPoint(id2, id1,  x,  y); return;}


  if (DISP)
    {
      printf("Recombination: %d\t%d\t\n", x, y);
      nodes[id1]->Display();
      nodes[id2]->Display();
    }

  if (DISP) printf("recombination step: %d\t%d [%d - %d]\n", id1, id2, x, y);

  if (x == 0) RecombinationPoint(id1, y + 1);
  if (y == nmarker-1) RecombinationPoint(id1, x);

  if (x != 0 && y + 1 < nmarker-1) // id1 <-- (n-2) (left) and (n-1) right. Then recombine n-1 at y
    {
      RecombinationPoint(id1, x);
      RecombinationPoint(n-1, y+1);
    }

}

void CFastARG::Display(void)
{
  Display(0);
}


void CFastARG::Display(int active)
{
  printf("\n\ncurrent stat\n");
  for (int i = 0; i != n; ++i)
    if (!active || nodes[i]->GetActive()) nodes[i]->Display();
}



void CFastARG::CheckCount(void)
{
  return;
  for (int l = 0; l != nmarker; ++l)
    {
      int t [10];
      for (int i = 0; i != 10; ++i) t[i] = 0;

      for (int i =0; i != n; ++i)
	if (nodes[i]->GetActive()) t[ nodes[i]->GetAl(l)] ++;

      for (int i =0; i != 3; ++i)
	if (t[i] != count[l*4+i]) 
	  {
	    printf("Error in counting: %d\n", i);
	    for (int j =0; j != 4; ++j)
	      printf("%d\t%d\n", count[l*4+j], t[j]);
	    exit(1);
	  }
    }
}


//Count distance between i and j
int CFastARG :: ReCountDist(int i, int j)
{
  char *s1  = nodes[i]->GetHap();
  char *s2  = nodes[j]->GetHap();

  int l = CountLeft(s1, s2);
  int l1 = 0;
  while ((s1[l1] == '3' || s2[l1] == '3') && l1 < l) l1++;
  return l - l1;
}



//Get an ARG tree
void CFastARG :: BuildARG(void)
{
  printf("start build ARG\n");

  srand (10);
  time_t start_run, end;
  time (&start_run);
  int nr = 0;
  while (nactivenodes > 1)
    {
      if (DISP) printf("\n\nstart new section\n");

      //      Display(1);

      if (nr++  % 100 == 0) {
	time (&end);
	printf("%d\t%d: [%.0f]\n", n, nactivenodes,  difftime(end, start_run));
      }
      //make mutation and coalescence
      Mutation();

      int mark[K10];
      for (int i =0; i != nactivenodes; ++i) mark[i] = 0;

      while (nactivenodes > 1)
	{
	  int maxdist = 0;
	  for (int ii = 0; ii != nactivenodes; ++ii)
	    if (!mark[ii])
	      {
		int i = activenodes[ii];
		int tdist = ReCountDist(i, pt[i].nei);

		if (tdist > maxdist) { que.Reset();   maxdist  = tdist;}
		if (tdist == maxdist) que.Push(ii);
	      }
	  if (maxdist == 0)
	    {
	      printf("error: %d\n", maxdist);
	      Display(1);
	      exit(1);
	    }
	  int lii = que.GetRandom();

	  if (lii == -1) continue;
	  mark[lii] = 1;

	  int id1 = activenodes[lii];
	  int id2 = pt[id1].nei;
	  RecombinationPoint(id1, id2, 0, pt[id1].dist-1);
	  break;
	}

    }
  BuiltARG = 1;
  printf("Finish Build fast ARG\n\n");
}




CARG_f*  CFastARG :: Export2CARG_f(int chr, int nmarker, int *marker, int nsample, char ** names)
{
  if (!BuiltARG) BuildARG();





  int *code = (int*)calloc(n+10, sizeof(int));
  int nc = 0;
  for (int i = 0; i != n; ++i)
    if (!nodes[i]->EdgeNode())  code[i] = nc++;



  CARG_f *p = new CARG_f;
  //  p->Init(H->GetChr(), H->GetNmarker(), H->GetPos(), H->GetNsample(), H->GetNames());
  p->Init(chr, nmarker, marker, nsample, names);

  //  printf("chr = %d\nnmarker = %d\n", chr, nmarker);

  //  printf("Export2CARG_f: Coming here\t%d\t%d\n", H->GetNmarker(),  H->GetNsample());

  for (int i = n-1; i > 0; --i)
    {
      if (nodes[i]->EdgeNode()) continue;
      if (nodes[i]->GetChild(1) == -1) continue;//leaves



      for (int k = 0; k != 2; ++k)
	{
	  int id = nodes[i]->GetChild(k+1);
	  if (id == -1) continue;

	  int st, ed;
	  if (!nodes[id]->EdgeNode())
	    {
	      st = nodes[i]->GetST();
	      ed = nodes[i]->GetED();
	    }
	  else
	    {
	      st = nodes[id]->GetST();
	      ed = nodes[id]->GetED();
	      id = nodes[id]->GetChild(1); 
	    }
	  //	  printf("i = %d, code[i] = %d\n", code[i], code[id]);

	  p->Add(code[i], code[id], st, ed);
	}
    }
  free(code);

  p->SetExist(1);
  return p;
}




CARG_f*  CFastARG :: Export2CARG_f(CHaplotypes *H)
{

  CARG_f *p = Export2CARG_f(H->GetChr(), H->GetNmarker(), H->GetPos(), 
			    H->GetNsample(),  H->GetNames());
  return p;
  /*
  if (!BuiltARG) BuildARG();

  printf("ok building ARG: n = %d\n", n	 );
  int *code = (int*)calloc(n+10, sizeof(int));
  int nc = 0;
  for (int i = 0; i != n; ++i)
    if (!nodes[i]->EdgeNode())  code[i] = nc++;

  CARG_f *p = new CARG_f;
  p->Init(H->GetChr(), H->GetNmarker(), H->GetPos(), H->GetNsample(), H->GetNames());

  //  printf("Export2CARG_f: Coming here\t%d\t%d\n", H->GetNmarker(),  H->GetNsample());

  for (int i = n-1; i > 0; --i)
    {
      //      printf("i = %d %d\n", i, n);
      if (nodes[i]->EdgeNode()) continue;
      if (nodes[i]->GetChild(1) == -1) continue;//leaves

      for (int k = 0; k != 2; ++k)
	{
	  int id = nodes[i]->GetChild(k+1);

	  if (id == -1) continue;

	  int st, ed;
	  if (!nodes[id]->EdgeNode())
	    {
	      st = nodes[i]->GetST();
	      ed = nodes[i]->GetED();
	  
	    }
	  else
	    {
	      st = nodes[id]->GetST();
	      ed = nodes[id]->GetED();
	      id = nodes[id]->GetChild(1); 
	    }
	  p->Add(code[i], code[id], st, ed);
	}
    }
  free(code);

  p->SetExist(1);
  return p;*/
}




CARG_f* CFastARG :: Build(CHap4ARG *H, char **sNames)
{
  printf("start building ARGs ...\n");

  int nsite =  H->GetNSite();
  int ns    =  H->GetNSample();
  int ichr  =  H->GetChrInt();//for 1 chromosome only
  printf("%d\t%d\t%d\n", nsite, ns, ichr);




  int *pos = (int*)calloc(nsite, sizeof(int));
  for (int i = 0; i != nsite; ++i) pos[i] = H->GetPos(i);
  char ** S =  H->Vert2FARG();


  //  for (int i = 1; i != nsite; ++i)     printf("%d\t%d\n", pos[i], pos[i] - pos[i-1]);  printf("\n\n");


  //  for (int i = 0; i != ns * 2; ++i) printf("%s\n", S[i]);


  Init(ns*2, nsite, S, pos, ichr);

  //  printf("coming to export\n");
  CARG_f* arg = Export2CARG_f(ichr, nsite, pos, ns, sNames);
   //  arg->Display();


  for (int i = 0; i != ns*2; ++i) free(S[i]);
  free(S);   //free S
  free(pos);   //  free pos

  printf("Building farg tree: Done\n");
  return arg;
}

