#include "COneSite.h"



COneSite::~COneSite()
{
  for (int i = 0; i != 5; ++i)
    if (CharArr[i] != NULL)       free(CharArr[i]);
  for (int i = 0; i != 5; ++i) CharArr[i] = NULL;
}


COneSite::COneSite()
{
  for (int i = 0; i != 5; ++i) CharArr[i] = NULL;

  gens = NULL; //genotypes
  genconf = NULL;//genotype confidence
  haps = NULL; //haplotypes
  hapconf = NULL; //haplotype confidence
  genphasedconf = NULL; //genotype confidences

  //  chr = -1;
  chr = NULL;
  pos = -1;
  conf = 1;
  len = 0;
}


int COneSite :: InitLength(int rlen)
{
  len = rlen;
  for (int i = 0; i != 5; ++i)
    if (i  < 3)
    {
      CharArr[i] = (char*)calloc(len + 1, sizeof(char));
      CharArr[i][len] = '\0';
    }
    else 
      {
	CharArr[i] = (char*)calloc(len/2 + 2, sizeof(char));
	CharArr[i][len/2] = '\0';
	CharArr[i][len/2+1] = '\0';
      }

  haps = CharArr[0];
  genconf = CharArr[1];
  genphasedconf = CharArr[4];
  
  gens = CharArr[2];
  hapconf = CharArr[3];
  return 1;
}

int COneSite :: MAF (void) //minor allele frequency
{
  int i = 0;
  if (haps == NULL) return 0;
  int len = strlen(haps);

  while (i < len && haps[i] == '?') i++;
  if (i == len) return 0;
  
  char al = haps[i];
  if (al >= 'a' && al <= 'z') al -= 32; //a-->z --> A --> Z


  int mas = 0; //minor allele samples
  for (int i = 0;  i != len/2; ++i)
    {
      if (haps[2*i] == '?' || haps[2*i+1] == '?') continue; //missing
      if (haps[i*2] != haps[i*2+1]) { mas ++; continue;} //hete
      if (haps[i*2] == al || haps[i*2]-32 == al)  continue;
      mas ++;
    }
  if (mas > len / 4)  mas = len/2 - mas;
  return mas;
}


void COneSite :: SetCharArr(char **s, int n)
{
  //len = strlength of the first sequence, haplotypes
  if (s[0] != NULL) len = strlen(s[0]);

  for (int i =0; i != 5; ++i)
    if (CharArr[i] != NULL) free(CharArr[i]);
  for (int i =0; i != 5; ++i) CharArr[i] = NULL; 

  for (int i = 0; i != n; ++i)
    if (s[i] != NULL)
    {
      CharArr[i] = (char*)calloc(strlen(s[i]) + 1, sizeof(char));
      strcpy(CharArr[i], s[i]);
    }

  haps = CharArr[0];
  genconf = CharArr[1];
  genphasedconf = CharArr[4];
  
  gens = CharArr[2];
  hapconf = CharArr[3];
}


COneSite* COneSite :: Clone(void)
{
  COneSite *p = new COneSite ();
  p->SetCharArr(CharArr, 5);
  p->SetConf(conf);
  p->SetChrPos(chr, pos);
  return p;
}



void COneSite::Vert2MargaritaFormat(void)
{
  for (int i =0; i != len; ++i)
    if (haps[i] == '?') haps[i] = '3';

  int i = 0;
  while (i < len && haps[i] == '3') i++;
  if (i == len) return;

  char al = haps[i];
  if (al >= 'a' && al <= 'z') al -= 32; //a -> A

  for (int i =0; i != len/2; ++i)
    {
      if (haps[i*2] == '3' || haps[i*2 + 1] == '3') continue;
      
      if (haps[i*2] != haps[2*i+1] && haps[2*i] <= 'Z') //unphased 
	{
	  haps[i*2] = '2';
	  haps[i*2] = '2';
	}

      for (int j = 0; j != 2; ++j)
      if (haps[i*2 + j] == al || haps[i*2+j]-32 == al) haps[i*2 + j] = '0';
      else haps[i*2+j] = '1';
    }
}


void COneSite :: Display(void)
{
  Display(0);
}



void COneSite :: Display(int type)
{
  if (type == 1&& haps == NULL) return;

  printf("%s\t%d: len = %d, %f\n", chr, pos, len, conf);
  if (conf < -0.0001) exit(1);
  char *names [5] = {"Haps", "HConf", "Gens", "GConf", "HGens"};
  for (int i = 0; i != 5; ++i)
    if (CharArr[i] != NULL) printf("%s\n%s\n", names[i], CharArr[i]);
  printf("\n\n");
}


void COneSite :: SetChrPos(int rchr, int rpos)
{
  char tempc[1024];
  sprintf(tempc, "%d", rchr);
  SetChrPos(tempc, rpos);

}



void COneSite :: SetChrPos(char *rchr, int rpos)
{
  if (chr != NULL) free(chr);
  chr = (char*)calloc(strlen(rchr)+1, sizeof(char));
  strcpy(chr, rchr);
  pos = rpos;
}



