#include "COneSiteLK.h"


COneSiteLK :: COneSiteLK()
{
  //  InitMemory(100);
  ms = 0;
  phred = NULL;
  depth = NULL;
  comments = NULL;

  pos = -1;
  nd = 0;
  ref = 'N';
  //  comments = (char*)calloc(10, sizeof(char));

  for (int i = 0; i != K1; ++i) samplecomments[i] = NULL;

  altall = NULL;
}






COneSiteLK :: ~COneSiteLK()
{
  if (phred != NULL) free(phred);
  if (depth != NULL) free(depth);
  if (altall != NULL) free(altall);
  if (comments != NULL) free(comments);
  for (int i = 0; i != K1; ++i)
    if (samplecomments[i] != NULL) { free(samplecomments[i]); samplecomments[i] = NULL;}
}

void COneSiteLK :: Reset(void)
{
  for (int i = 0; i != ms * NG;  ++i) phred[i] = 0;
  for (int i = 0; i != ms; ++i) depth[i] = 0;

  for (int i = 0; i != K1; ++i)
    if (samplecomments[i] != NULL) { free(samplecomments[i]); samplecomments[i] = NULL;}


  nd = 0;
  pos = -1;
  posend = -1;
  ref = 'N';
  alt = '.';

  if (comments != NULL) free(comments); 
  if (altall != NULL) free(altall);

  comments = NULL;
  altall = NULL;
}

void COneSiteLK :: Realloc(int size)
{
  int oms = ms;
  ms = size;
  phred = (unsigned char*) realloc(phred  , ms * NG * sizeof(unsigned char));
  depth = (int*)realloc(depth, ms * sizeof(int)); 

  for (int i = oms; i != ms; ++i) 
    for (int j = 0; j != NG; ++j) phred[i*NG + j] = 0;
  for (int i = oms; i != ms; ++i) depth[i] = 0;
  

}

void COneSiteLK :: InitSize(int size)
{
  if (phred != NULL) free(phred);
  if (depth != NULL) free(depth);
  phred = NULL;
  depth = NULL;
  Realloc(size);

}

void COneSiteLK :: InitSampleName(int i_nsn, char ** i_samplenames)
{
  nsn = i_nsn;
  samplenames = i_samplenames;
  InitSize(nsn);
}

int COneSiteLK :: Add(COneLK *s, int indel)
{ 
  if (pos == -1) { 
    pos = s->GetPos();
    strcpy(chr, s->GetChr());
    if (altall != NULL) free(altall);
    if (indel) altall = Clone(s->GetAltAll());
    if (posend == -1) posend = s->GetPosEnd();
  }

  if (s->GetPos() != pos || strcmp(s->GetChr(), chr) != 0) return NEWSITE; //

  if (indel)
    {
      if (s->GetPosEnd() != posend) {
	fprintf(stderr, "indels start at the same location (%s:%d)but end at different location [%d, %d]\n", s->GetChr(), s->GetPos(),  posend,  s->GetPosEnd()  );
	return NEWSITE;
      }

      if (altall != NULL && s->GetAltAll() != NULL) 
	if (strcmp(altall, s->GetAltAll()) != 0) return NEWSITE;
    }

  if (ref == 'N') ref = s->GetRef();
  unsigned char *p = s->GetLK();
  char *sn = s->GetName();
  int id = -1;

  if (nsn == 0)       id = nd; //no data
  else
    for (int i =0; i != nsn; ++i)
      if (strcmp(sn, samplenames[i]) == 0) { id = i; break;}

  if (id != -1)
    {
      while (id >= ms)  Realloc(ms + 100); 
      for (int i =0; i != NG; ++i)	phred[id * NG + i] = p[i];
      depth[id] = s->GetDepth();

      if (samplecomments[id] != NULL) free(samplecomments[id]);
      samplecomments[id] = Clone(s->GetComment());
      nd ++;
    }


  return OK;
}


void COneSiteLK::VCF_Write(FILE *fo, char *comment, int indel, int *index)
{
  if (chr[0] == '0' && chr[1] <= '9' && chr[1] > '0' && strlen(chr) == 2)
    fprintf(fo, "%d\t%d", atoi(chr), pos);
  else fprintf(fo, "%s\t%d", chr, pos);


  fprintf(fo, "\t."); //dbsnp
  if (!indel)
    {
      fprintf(fo, "\t%c", ref);
      if (index != NULL)
	{
	  fprintf(fo, "\t");
	  int nalt = 0;
	  for (int i = 0; i != 255; ++i) 
	    if (index[i] > '0') { fprintf(fo, "%c", i); nalt ++;}
	  if (nalt == 0) fprintf(fo, ".");
	}
      else
	{
	  fprintf(fo, "\t%c", alt);
	  if (ref == 'N') 
	    fprintf(fo, ",%c", mref);
	}
    }
  else {
    fprintf(fo, "\t.");
    if (altall != NULL) fprintf(fo, "\t%s", altall);
    else fprintf(fo, "\t.");
  }
  //  printf("conf = %f\n", conf);

  fprintf(fo,"\t%d", Prob2Phred(1-conf));

  if (conf < .9) fprintf(fo, "\tq10");
  else fprintf(fo, "\t0");

  int dep = 0;
  if (nsn > 0) 
    for (int i = 0; i != nsn; ++i) dep += depth[i];
  else 
    for (int i = 0; i != nd; ++i) dep += depth[i];


  fprintf(fo, "\tNS=%d", nd);
  if (!indel) fprintf(fo, ";DP=%d", dep);

  
  if (comments != NULL) fprintf(fo, ";%s", comments);

  if (comment != NULL)  fprintf(fo, ";%s", comment);


  if (!indel)    fprintf(fo, "\tGT:GQ:DP:phred");
  else    fprintf(fo, "\tGT:GQ");

  if (comment != NULL) 
    if (strstr(comment, "Phasing") != NULL)
      fprintf(fo, ":HQ1,HQ2");

}


/*
double*VCFOneSiteLK :: GetLogProb(void)
{

  int nt = GetND();
  double *out = (double*)calloc(nt, sizeof(double));
  for (int i = 0; i != nt; ++i)  out[i] = phred[i];
  return out;
}


unsigned char **  COneSiteLK :: GetPhred(void)
{
  int nt = GetND();
  unsigned char **  *out = (unsigned char **)calloc(nt, sizeof(unsigned char*));
  for (int i = 0; i != nt; ++i)  out[i] = phred[i];
  return out;
}
*/

int COneSiteLK :: GetRefID(void)
{
  char DNA[] = "ACGT";
  for (int i =0; i != 4; ++i)
    if (ref == DNA[i]) return i;
  return -1;
}



void COneSiteLK :: Genotype(char *haps)
{

  //prob(g|d) = 10^(-phred/10) * prob
  //=10^((-phred + 10*log(prob)) / 10)


  double log10p01 = -20;//10 * log(0.01) / log (10);
  double log10p99 = -0.04364805402;//10 * log(0.99) / log (10);

  char DNA[] = "ACGT";


  for (int i = 0; i != nsn; ++i)
    {
      unsigned char *p = &phred[i*NG];

      int l = 0;
      double maxt = -(1E+10);

      for (int a = 0; a != NA; ++a)
	for (int b = a; b != NA; ++b)
	  {
	    double t = (double) p[l++];
	    if (a == b) t = -t +  log10p99;
	    else t = -t + log10p01;

	    if (t > maxt) {
	      haps[i*2]   = DNA[a];
	      haps[i*2+1] = DNA[b];
	      maxt = t;
	    }
	    //	    printf("%s\t%d\t: %c%c: %d %f\n", chr, pos, DNA[a], DNA[b], p[l-1], t);

	  }
      //      printf("\n");
    }
  haps[2*nsn] = '\0'; 
}


void COneSiteLK :: SetComments(char *s)
{
  if (comments == NULL) comments = Clone(s);
  else
    {
      comments = (char*)realloc(comments, (strlen(comments) + strlen(s) + 10) * sizeof(char));
      strcat(comments, s);

    }
  //  fprintf(stderr, "%s\n", comments);

}


int COneSiteLK :: CheckSNPConf(void)
{

  char DNA[] = "ACGT";

  int refid = -1;
  for (int i = 0; i != 4; ++i)
    if (DNA[i] == ref || DNA[i] == ref - 32) refid = i;

  if (refid == -1) return 0; //no reference


  int ns;
  if (nsn == 0) ns = nd;
  else ns = nsn;

  int sum = 0;
  for (int i = 0; i != ns; ++i)
    sum += phred[i*NG + refid];

  if (sum < 5) return 0;

  return 1;
}



void COneSiteLK :: Display(void)
{
  int ns;
  if (nsn == 0) ns = nd;
  else ns = nsn;


  printf("\n---------------\n%s\t%d\t%d\n", chr, pos, nd);
  for (int i = 0; i != ns; ++i)
    {
      printf("%d:\t", i);
      for (int j =0; j != NG; ++j) printf("%d\t", phred[i*NG + j]);
      if (nsn > 0) printf("%s", samplenames[i]);
      printf("\n");
    }
}

