#include "COutput.h"

COutput :: COutput()
{
  strcpy(fnout, "");
  //  fo = stdout;
  fo = NULL;
  indel = 0;
  int l = 0;
  for (int a = 0; a != NA; ++a)
    for (int b = a; b != NA; ++b)
      {
	order[a][b] = l;
	order[b][a] = l;
	l++;
      }
  
  strcpy(DNA, "ACGT");
  double log10 = log(10);
  for (int i = 0; i != K1; ++i)
    {
      double x = i;
      Phred2Prob[i] = exp(-x/10 *log10);
    }

  SetPriorProbHom(0.99);
  header = 0;
}

COutput :: ~COutput()
{
  if (fo != NULL) fclose(fo);
}

void COutput :: Init(char *fn) 
{ 
  if (strcmp(fn, "-") != 0) //stdout
    {
      strcpy(fnout, fn);
      fo = fopen(fnout, "w");

      if (fo == NULL)
	{
	  fprintf(stderr, "error while open to write: %s\n", fn);
	  exit(1);
	}
    }
  else fo = stdout;
}


void COutput :: Output(COneSiteLK *s)
{
  s->VCF_Write(fo);


  if (s->GetNsn() == 0) {
    fprintf(fo, "\n");
    return;
  }


  unsigned char *phredlk = s->GetLK();
  int *depth = s->GetDepth();
  double prob[NG];


  //  int refid = s->GetRefID();
  
  int rid = order[s->GetRefID()][s->GetRefID()];


  int ia, iab;
  int minsum = 10000000;

  for (int a = 0; a != NA; ++a)
    if (a != s->GetRefID())
    {
      int ab = order[s->GetRefID()][a];

      int sum = 0;
      for (int i = 0; i != nsn; ++i)
	if (depth[i] != 0)
	  sum += MIN3( phredlk[ LOC(i, rid)],  phredlk[ LOC(i, order[a][a])], phredlk[LOC(i, ab)]);
      
      if (sum < minsum) { 
	ia =  order[a][a];
	iab = ab;
	minsum = sum;
      }
    }

  //  if (s->GetPos() == 44004210) printf("ia = %d, iab = %d, refid = %d\n", ia, iab, rid);

  for (int i = 0; i != nsn; ++i)
    if (depth[i] == 0) fprintf(fo, "\t./.:-1:0");//no data
    else
      {
	for (int j = 0; j != NG; ++j)  prob[j] = Phred2Prob[ phredlk[  LOC(i, j) ] ];

	for (int a = 0; a != NA; ++a)
	  for (int b = a+1; b != NA; ++b) prob[ order[a][b] ] *= pphet;

	for (int a = 0; a != NA; ++a)	 prob[ order[a][a] ] *= pphom;
	Norm(prob, NG);


	double t = MAX3(prob[rid], prob[ia], prob[iab]);

	//	if (s->GetPos() == 44004210) printf("ia = %d, iab = %d, refid = %d, t = %f, prob[rid] = %f\n", ia, iab, rid, t, prob[rid]);

	if (t == prob[rid]) fprintf(fo, "\t0/0");
	else 
	  if (t == prob[ia])   fprintf(fo, "\t1/1");
	  else fprintf(fo, "\t1/0");
	
	fprintf(fo, ":%d:%d",  Prob2Phred(1-t), depth[i] ); 
	fprintf(fo, ":%d,%d,%d", phredlk[ LOC(i, rid)], phredlk[LOC(i, ia)], phredlk[LOC(i, iab)]);

      }
  fprintf(fo, "\n");
}



/*
Output:
    20      9139    .       .      D3    100     0       NS=52   
    GT:GQ:HQ:CM    1|0:41:43,102   0/0:123::my_data   1/1:112
*/

void COutput :: OutPut(COneSite *s, COneSiteLK *os)
{

  //  printf("coming here\n");
  //  os->VCF_Write(fo);

  int *depth = os->GetDepth();
  char ** chrarr = s->GetCharArr();

  char *gens    = chrarr[2];
  char *genconf = chrarr[3];

  char *haps    = chrarr[0];
  char *hapconf = chrarr[1];
  char *genphaseconf = chrarr[4];


  int *index = GetIndex(haps, os->GetRef());

  os->VCF_Write(fo, "Phasing", indel, index);


  for (int i = 0; i != nsn; ++i)
    {
      fprintf(fo, "\t%c|%c:%d", index[haps[i*2]], index[haps[i*2+1]], genphaseconf[i]);
      if (!indel)
	fprintf(fo, ":%d", depth[i]);

      fprintf(fo, ":%d,%d", hapconf[i*2], hapconf[i*2+1]);
      if (indel)
	if (os->GetComment(i) != NULL) fprintf(fo, "%s", os->GetComment(i));

    }
  fprintf(fo, "\n");
  free(index);


  index = GetIndex(gens, os->GetRef());
  os->VCF_Write(fo, "Genotyping", indel, index);
  for (int i = 0; i != nsn; ++i)
    {
      fprintf(fo, "\t%c/%c:%d", index[gens[i*2]], index[gens[i*2+1]], genconf[i]);
      if (!indel)
	fprintf(fo, ":%d",depth[i]);

      if (indel)
	if (os->GetComment(i) != NULL) fprintf(fo, "%s", os->GetComment(i));
    }
  fprintf(fo, "\n");

  //  printf("out here\n");

}

void COutput:: VCF_Write_Header(int phased)
{
  time_t rawtime;
  time (&rawtime);
  fprintf(fo, "##fileformat=VCFv3.3\n");
  fprintf(fo, "##fileDate=%s", ctime (&rawtime) );
  fprintf(fo, "##source=QCall\n");
  fprintf(fo, "##phasing=NoPhase\n");
  fprintf(fo, "##INFO=NS,1,Integer,\"Number of Samples With Data\"\n");
  fprintf(fo, "##INFO=DP,1,Integer,\"Total Depth\"\n");
  fprintf(fo, "##INFO=AF,-1,Float,\"Allele Frequency\"\n");
  fprintf(fo, "##INFO=AA,1,String,\"Ancestral Allele\"\n");
  fprintf(fo, "##FILTER=q10,\"Quality below 10\"\n");
  fprintf(fo, "##FILTER=s50,\"Less than 50%% of samples have data\"\n");
  fprintf(fo, "##FORMAT=GT,1,String,\"Genotype\"\n");
  fprintf(fo, "##FORMAT=GQ,1,Integer,\"Genotype Quality\"\n");
  fprintf(fo, "##FORMAT=DP,1,Integer,\"Read Depth\"\n");

  if (phased) 
    fprintf(fo, "##FORMAT=HQ,2,Integer,\"Haplotype quality\"\n");

  fprintf(fo, "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT");
  for (int i =0; i != nsn; ++i)    fprintf(fo, "\t%s", samplenames[i]);
  fprintf(fo, "\n");
}


