#include "CReadDataOneLK.h"

CReadDataOneLK::CReadDataOneLK()
{
  nlk = 0;
  mlk = 0;
  cid = 0;

  lkdata = NULL;  
  vcf    = NULL;
  snpdec = NULL;
  snpcan = NULL;

  double log10 = log(10);
  for (int i = 0; i != K1; ++i)
    {
      double x = i;
      Phred2LK[i] = exp(-x/10 *log10);
    }
  freefin = 0;

  strcpy(DNA, "ACGT");
  for (int i = 0; i != 4; ++i)    OrderDNA[ DNA[i]]  = i;
  for (int i = 0; i != 4; ++i)    OrderDNA[ DNA[i] + 32]  = i;

  int l = 0;
  for (int a = 0; a != NA; ++a)
    for (int b = a; b != NA; ++b)
      {
	order[a][b] = l;
	order[b][a] = l;
	l++;
      }

  samplenames = (char**)calloc(K10, sizeof(char*));
  nsn = 0;

  strcpy(fnoutcan, "");
  fco = NULL;
  mlk = K10;
  lkdata = (COneLK**) malloc(mlk * sizeof(COneLK*));  
  keepdata = 1;

  tempfile = 1;
  pphet = 0.01;
  pphom = 0.99;
}

CReadDataOneLK::~CReadDataOneLK()
{
  FixedVCFFile();

  if (nlk > 0)
    for (int i = 0; i != nlk; ++i) delete lkdata[i];
  if (lkdata != NULL) free (lkdata);

  if (snpcan != NULL) delete snpcan;
  if (vcf != NULL) delete vcf;
  if (freefin) free(f);
  
  for (int i = 0; i != nsn; ++i) free(samplenames[i]);
  free(samplenames);
  if (fco != NULL) fclose(fco);
}


void CReadDataOneLK :: PushNames(char *s)
{
  if (s == NULL) return;
  for (int i = 0; i != nsn; ++i) 
    if (strcmp(s, samplenames[i]) == 0) return;

  samplenames[nsn] = (char*)calloc(strlen(s) + 1, sizeof(char));
  strcpy(samplenames[nsn++], s);
}


int CReadDataOneLK :: SNPCanFilter(OneSite *s)
{
  if (snpdec == NULL) return 0;

  if (s->ndata < 3) return 1;//remove 

  double *prob =  GetLK(s);
  double conf, maf;
  char altall;

  if (DEBUG) printf("computing: %s\t%d\n", s->data[0]->GetChr(), s->data[0]->GetPos());

  snpdec->SNPConfidence(prob, s->ndata, OrderDNA[s->data[0]->GetRef()], conf, altall, maf);
  free(prob);

  if (conf < snpdec->GetThreshold())  return 1; //filtered by SNPCAN
  if (fco == NULL) return 0; //just keep
      

  if (DEBUG)  printf("%d: %d %s\t%d\t%f\n",nlk, mlk,   s->data[0]->GetChr(), s->data[0]->GetPos(), conf);

  VCFSNPCAN *p = new VCFSNPCAN;
  p->SetChrPos(s->data[0]->GetChr(), s->data[0]->GetPos());
  p->SetConf(conf);
  p->SetNData(s->ndata);
  p->SetDepth(TotalDepth(s) );
  p->SetAltAll(altall);
  p->SetMAF(maf);
  p->SetRef(s->data[0]->GetRef());

  VCF_Write_OneSite(fco, p, s);
  delete p;
  return 0;
}

void CReadDataOneLK :: PushOneSite(COneLK *p)
{
  if (nlk >= mlk)
    {
      mlk += K10;
      lkdata = (COneLK**) realloc(lkdata, mlk * sizeof(COneLK*));

      if (lkdata == NULL)
	{
	  printf("error in reaclloc memory for data: mlk = %d\n", mlk);
	  exit(1);
	}
    }
  lkdata[nlk++] = p;

}

void CReadDataOneLK::Read(int nr)
{
  int cid = 0;

  while (!feof(f))
    {
      COneLK *p = new COneLK;
      if (p == NULL)
	{
	  fprintf(stderr, "Overload Memory:\n");
	  break;
	}

      if (!p->Read(f)) { delete p; continue;}
      //      p->Display();
      PushNames(p->GetName());


      if (vcf != NULL)
	{
	  vcf->Locate(p->GetChr(), p->GetPos());
	  if (!p->Ident(vcf->GetChrS(),  vcf->GetPos() ))
	    {
	      delete p;
	      continue;
	    }
	}

      if (nlk == 0) { PushOneSite(p);  continue;} //first data
      if (lkdata[nlk-1]->Ident(p)) {PushOneSite(p);  continue;}//same chr, pos         


      OneSite site;
      site.data = &lkdata[cid]; 
      site.ndata = nlk - cid;
      //      Display(&site);


      int t = 0;
      if (snpdec != NULL) t = SNPCanFilter(&site);

      if (t == 1 || !keepdata )
	{
	  for (int i = cid; i < nlk; ++i) delete lkdata[i];
	  nlk = cid;
	}
      cid = nlk;
      PushOneSite(p);
    }

  OneSite site;
  site.data = &lkdata[cid]; 
  site.ndata = nlk - cid;
  if (nlk > 0 && snpdec != NULL) SNPCanFilter(&site);
}

void CReadDataOneLK::Free(void)
{
  if (cid < M1) return;
  for (int i = 0; i != cid; ++i) delete lkdata[i];
  int l = 0;
  for (int i = cid; i != nlk; ++i) lkdata[l++] = lkdata[i];
  nlk = l;
  cid = 0;
}




COneLK ** CReadDataOneLK :: GetSite(int &n)
{
  if (cid + K10 > nlk) Read(K100);
  
  if (cid == nlk) return NULL;

  char *s = lkdata[cid]->GetChr();
  int pos = lkdata[cid]->GetPos();

  n = 0;
  for (int i = cid; i != nlk; ++i)
    if (lkdata[i]->Ident(s, pos)) n++;
    else break;


  COneLK ** p = (COneLK**)calloc(n, sizeof(COneLK*));
  for (int i = 0; i != n; ++i)  p[i] = lkdata[cid++];
  return p;
}


void CReadDataOneLK::Test(void)
{
  int n;
  while (1==1)
    {
      COneLK ** p = GetSite(n);
      if (p == NULL) return;
      free(p);
      Free();
    }
}


double *CReadDataOneLK :: GetLK(OneSite *s)
{
  double *out = (double*)calloc(s->ndata * NG, sizeof(double));

  for (int i = 0; i != s->ndata; ++i)
    {
      double *p = &out[i*NG];
      unsigned char *lk = s->data[i]->GetLK();

      for (int j = 0; j != NG; ++j)
	{
	  int x = lk[j];
	  p[j] = Phred2LK[x];
	}
    }
  return out;
}

/*
double * CReadDataOneLK :: GetLK(int st, int ed)
{
  double *out = (double*)calloc( (ed - st + 1)*NG, sizeof(double));
  for (int i = st; i != ed + 1; ++i) 
    GetLK(lkdata[i], &out[ (i-st) * NG]);
  return out;
}
*/

void CReadDataOneLK :: GetLK(COneLK *data, double *out)
{
  unsigned char   *lk = data->GetLK();    
  for (int j = 0; j != NG; ++j) {                
    int x = lk[j];   
    out[j] = Phred2LK[x];        
  }
}


int CReadDataOneLK :: TotalDepth(OneSite *s)
{
  int total = 0;
  for (int i = 0; i < s->ndata; ++i) total += s->data[i]->GetDepth();
  return total;
}

void CReadDataOneLK::Init(char *fn)
{
  f = fopen(fn, "r");
  if (f == NULL)
    {
      printf("Error opening file: %s\n", fn);
      exit(1);
    }
  freefin = 1;
}

void CReadDataOneLK :: SetVCF(char *fn)
{
  if (vcf == NULL) 
    {
      vcf = new CVCF;
      vcf->Init(fn);
    }
}


void CReadDataOneLK :: Display(void)
{
  if (f == NULL) return;
  while (!feof(f))
    {
      Read();
      for (int i =0; i != nlk; ++i) lkdata[i]->Display();
    }
}


void CReadDataOneLK :: VCF_Write_Header(FILE *fco, int phased)
{
  time_t rawtime;
  time (&rawtime);

  fprintf(fco, "##fileformat=VCFv3.3\n");
  fprintf(fco, "##fileDate=%s", ctime (&rawtime) );
  fprintf(fco, "##source=QCall\n");
  fprintf(fco, "##phasing=NoPhase\n");
  fprintf(fco, "##INFO=NS,1,Integer,\"Number of Samples With Data\"\n");
  fprintf(fco, "##INFO=DP,1,Integer,\"Total Depth\"\n");
  fprintf(fco, "##INFO=AF,-1,Float,\"Allele Frequency\"\n");
  fprintf(fco, "##INFO=AA,1,String,\"Ancestral Allele\"\n");
  fprintf(fco, "##FILTER=q10,\"Quality below 10\"\n");
  fprintf(fco, "##FILTER=s50,\"Less than 50%% of samples have data\"\n");
  fprintf(fco, "##FORMAT=GT,1,String,\"Genotype\"\n");
  fprintf(fco, "##FORMAT=GQ,1,Integer,\"Genotype Quality\"\n");
  fprintf(fco, "##FORMAT=DP,1,Integer,\"Read Depth\"\n");

  if (phased) 
    fprintf(fco, "##FORMAT=HQ,2,Integer,\"Haplotype quality\"\n");
  fprintf(fco, "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT");
  for (int i =0; i != nsn; ++i)    fprintf(fco, "\t%s", samplenames[i]);
  fprintf(fco, "\n");
}


void CReadDataOneLK :: VCF_Write_SNPCan(void)
{
  return ;
}






void CReadDataOneLK :: VCF_Write_OneSite(FILE *fco, VCFSNPCAN *site, OneSite *s)// COneLK** data, int ns)
{
  //  char *chr = site->GetChr();
  //  int  pos = site->GetPos();

  site->VCF_Write(fco, s->ndata);

  COneLK** data = s->data;
  int ndata = s->ndata;

  double p[NG+10];
  for (int j = 0; j != nsn; ++j) 
    {
      int depth = 0;
      for (int k = 0; k < ndata; ++k)
	if (strcmp(data[k]->GetName(), samplenames[j]) == 0) {
	  
	  GetLK(data[k], p);

	  depth = lkdata[k]->GetDepth();
	  break;
	}
      if (depth == 0)//no data
	fprintf(fco, "\t./.:-1:0");
      else
	{
	  for (int a = 0; a != NA; ++a)
	    for (int b = a+1; b != NA; ++b)	  p[ order[a][b]] *= pphet;

	  for (int a = 0; a != NA; ++a)	 p[order[a][a]] *= pphom;

	  Norm(p, NG);

	  double maxp = -1;
	  int ma, mb;
	  for (int a = 0; a != NA; ++a)
	    for (int b = a; b != NA; ++b)
	      if (p[order[a][b]] > maxp) {
		maxp = p[order[a][b]];
		ma = a;
		mb = b;
	      }
	  fprintf(fco, "\t%c/%c:%d:%d", DNA[ma], DNA[mb], Prob2Phred(1-maxp), depth);
	}
    }
  fprintf(fco, "\n");
}


void CReadDataOneLK :: SetSNPCanOut(char *fn)
{
  strcpy(fnoutcan, fn);
  if (tempfile)
    {
      char tn[1024];
      sprintf(tn, "%s_temp.txt", fn);
      fco = fopen(tn, "w");
      VCF_Write_Header(fco, 0);
    }
  else
    {
      fco = fopen(fn, "w");
      VCF_Write_Header(fco, 0);
    }
}



void CReadDataOneLK :: FixedVCFFile(void)
{
  if (!tempfile) return;

  if (fco == NULL) return; //
  fclose(fco);			// 
  fco = NULL;

  if (DEBUG)  for (int i =0; i != nsn; ++i) printf("%s\n", samplenames[i]);
  int id [K10];
  for (int i = 0; i != nsn; ++i) id[i] = i;

  for (int i = 0; i != nsn; ++i)
    for (int j = i + 1; j != nsn; ++j)
      if (strcmp(samplenames[ id[i] ], samplenames[ id[j]]) > 0) {
	int tempi = id[i];
	id[i] = id[j];
	id[j] = tempi;
      }
  
  char *bf = (char*)calloc(K100, sizeof(char));
  FILE *fo = fopen(fnoutcan, "w");

  char tn[1024];
  sprintf(tn, "%s_temp.txt", fnoutcan);
  FILE *fin = fopen(tn, "r");
  while (!feof(fin))
    {
      fgets(bf, K100, fin);
      if (strstr(bf, "#CHROM") != NULL) break;
      fprintf(fo, "%s", bf);
    }

  int len = strlen(bf)-1;
  while (len > 0 && bf[len] < 33) len --;
  bf[len + 1] = '\0';
  fprintf(fo, "%s", bf);
  for (int i = 0; i != nsn; ++i) fprintf(fo, "\t%s", samplenames[ id[i]]);~

  fprintf(fo, "\n");

  char ** pdata = (char**)calloc(nsn + 10, sizeof(char*));
  while (!feof(fin))
    {
      strcpy(bf, "");
      fgets(bf, K100, fin);   
      CleanBF(bf);

      if (strlen(bf) < 1) continue;

      int nc = 0;
      //      char *p = bf;
      int i = 0, len = strlen(bf);

      while (i < len && nc < 9)
	if (bf[i++] < 33) nc ++;
      bf[i-1] = '\0';
      fprintf(fo, "%s", bf);


      int l = 0;
      pdata[l++] = &bf[i];

      while (i < len) 
	{
	  if (bf[i] == '\t') {
	    pdata[l++] =&bf[i+1]; bf[i] = '\0';
	  }
	  i++;
	  
	}
      
      //      for (int i =0; i != l; ++i) printf("i  = %d, daat = %s\n", i, pdata[i]);

      for (int k = 0; k != nsn; ++k)
	if (id[k] >= l) fprintf(fo, "\t./.:-1:0");
	else fprintf(fo, "\t%s", pdata[ id[k]]);
      fprintf(fo, "\n");
    }

  fclose(fo);
  fclose(fin);
  free(bf);
  free(pdata);

  char cmd[1024];
  sprintf(cmd, "rm %s", tn);
  system(cmd);
}


void  CReadDataOneLK :: Display(OneSite *s)
{
  printf("%d\n", s->ndata);
  for (int i =0; i != s->ndata; ++i) s->data[i]->Display();
  printf("\n\n");

}


void CReadDataOneLK :: SetSampleNames(char *fn)
{
  FILE *f = fopen(fn, "r");
  if (f == NULL) 
    {
      printf("error open files of samples: %s\n", fn);
      exit(1);
    }

  nsn = 0;

  char tempc[1024];
  while (!feof(f))
    {
      if (fscanf(f, "%s", tempc) != 1) continue;
      samplenames[nsn] = (char*)calloc(strlen(tempc) + 1, sizeof(char));
      strcpy(samplenames[nsn++], tempc);
    }
  fclose(f);

  tempfile = 0;
  SetSNPCanOut(fnoutcan);


}




void CReadDataOneLK :: SetPPHet(double x)
{
  if (x > 1 || x < 0)
    {
      printf("pphet must be a number in [0..1]\n");
      exit(1);
    }

  pphet = x;
  pphom = 1-x;
}
