#include "CTree.h"

CTree::CTree()
{
  T = NULL;
  numfigs = 0;
  for (int i = 0; i != MAXSAMPLE; ++i) logprob[i] = NULL; 

  logConf = NULL;
  maxlogProb = 0;
}


CTree::~CTree()
{
  for (int i = 0; i != numfigs; ++i) free(figuration[i]);
  if (T != NULL) delete [] T;
  if (logConf != NULL) FreeDouble(logConf, numfigs);
}



int CTree::CheckFig(char *s)
{

  char s1[K10];
  int len = strlen(s);
  for (int i =0; i != len; ++i)
    s1[i] = 1  - (s[i] - '0') + '0';

  for (int i = 0; i != numfigs; ++i)
    {
      if (strcmp(s, figuration[i]) == 0) return 1;
      if (strcmp(s1, figuration[i]) == 0) return 1;
    }

  printf("%s\n", s);
  for (int i =0; i != numfigs; ++i)
    {
      for (int j = 0; j != len; ++j)
	if (s[j] == figuration[i][j]) printf(".");
	else printf("%c", figuration[i][j]);

      printf("\n");
    }

  return 0;
}

void CTree :: ViewL(int id, char *s)
{
  if (T[id].left == -1 && T[id].right == -1) 
    {
      if (s == NULL)
	printf("%d", id);
      else 
	printf("%c", s[id]);
      return ;//leaf
    }

  printf("(");  ViewL(T[id].left, s);
  printf(",");
  ViewL(T[id].right, s);  printf(")");

}

void CTree :: ViewL(char *s){ 
  ViewL(rid, s);
  printf(";\n");
}

void CTree :: View (int id)
{
  View(id, NULL);
}

//view node id with label from s

void CTree::View(int id, char **s)
{
  if (T[id].left == -1 && T[id].right == -1) 
    {
      if (s == NULL)
	printf("%d", id);
      else 
	printf("%s", s[id/2]);
      return ;//leaf
    }

  printf("(");  View(T[id].left, s);
  printf(",");
  View(T[id].right, s);  printf(")");
}



void CTree::View(void)
{
  printf("%d\t%d\n", chr, pos);

  View(rid, NULL);
  printf(";\n");
}


void CTree::View(char** s)
{
  View(rid, s);
  printf(";\n");
}


//return a vector of nleaf with values 0
char* CTree::InitLeaves(void)
{
  char *t = (char*)calloc(nleaf+10, sizeof(char));
  for (long j = 0; j != nleaf; ++j) t[j] = '0';
  t[nleaf] = '\0';
  return t;
}


//calculate prior probability of figuration

void CTree::CalPriorProp(char** conf, int numfigs)
{
  int count[K10];
  for (int i =0; i != nleaf; ++i) count[i] = 0;

  int Ind[K10];
  double pp[K10];

  for (int i = 0;  i != numfigs; ++i)
    {
      char *p = conf[i];
      int nc = 0;
      for (int j = 0; j != nleaf; ++j)
	if (p[j] == p[0]) nc++;
      
      if (nc > nleaf/2)  nc = nleaf - nc;
      count[nc] ++;
      Ind[i] = nc;
    }

  for (int i = 1; i != nleaf; ++i)
    if (count[i] > 0) pp[i] = 1/(double) i + 1/(double) (nleaf -i); 
    else pp[i] = 0;
  pp[0]  = 0;

  double sum = 0;
  for (int i = 1; i != nleaf; ++i)    sum += pp[i];
  for (int i = 1; i != nleaf; ++i)    pp[i] /= sum;
  for (int i = 1; i != nleaf; ++i)   
    if (count[i] > 0) pp[i] /= (double) count[i];
  
  pp[0] = 0;
  pp[nleaf] = 0;
  for (int i = 0; i != numfigs; ++i) PriorP[i] = pp[ Ind[i]];
  //  for (int i = 0; i != numfigs; ++i) printf("%s\t%f\n", conf[i], PriorP[i]);
}


//Set One Mutation
//Build figurations caused by one mutation
void CTree::SetOneMut(char** out)
{
  numfigs = 0;
  for (int i = nleaf; i != 2*nleaf-1; ++i) //number of node, inside nodes
    {
      out[numfigs] = InitLeaves();
      Put(T[i].right, 1, out[numfigs++]); 

      out[numfigs] = InitLeaves();
      Put(T[i].left, 1, out[numfigs++]);
    }
  CalPriorProp(out, numfigs);
}


//Set node nid = v and is
void CTree::Put(int nid, char v, char* g)
{
  if (T[nid].left == -1 && T[nid].right == -1) //leaf
    {
      if (nid >= nleaf) {
	printf("big problem in tree: %d %d\n", nid, nleaf);
	View();
	exit(1);
      }
      g[nid] = v + '0';
    }
  else
    {
      Put(T[nid].left,  v, g);
      Put(T[nid].right, v, g);
    }
}

//just view all figurations
void CTree::ViewFig(void)
{
  printf("\n");
  for (int i = 0;  i != numfigs; ++i)
    printf("%s\n", figuration[i]);
}

int CTree::Init(FILE *f)
{
  int rchr, rpos;
  char bf[M1];
  int pint[K10];

  if (fscanf(f, "%d %d", &rchr, &rpos) != 2) return 0;
  fgets(bf, M1, f);

  int np = SplitInt(bf, pint);

  //  printf("%s, %d\n", bf, np);
  if (np == 0) return 0;

  if (np % 3 != 0)
    {
      printf("error: np = %d\n%d\t%d%s\n", np, chr, pos, bf);
      exit(1);
    }

  nleaf = np / 3 + 1;
  Init(pint, nleaf, rchr, rpos);
  
  return 1;
}

//input data from a vector pint, rnleaf, <rchr, rpos>
int CTree::Init(int *pint, int rnleaf, int rchr, int rpos)
{
  chr = rchr;
  pos = rpos;
  nleaf = rnleaf;
  nsp = nleaf / 2 ;
  //  nsample = nleaf / 2 ;
  if (T != NULL) delete [] T;
  T = new Node[nleaf * 2 +10]; //locate node
  
  for (int i = 0; i != nleaf * 2 + 10; ++i){
    T[i].left = -1;
    T[i].right = -1;
    T[i].par = -1;
  }

  for (long i = 0; i != nleaf-1; ++i)
    {
      int child1  = pint[i*3];
      int child2  = pint[i*3 + 1];

      int par     = pint[i*3 + 2];

      T[par].left = child1;
      T[par].right = child2;

      T[child1].par = par;
      T[child2].par = par;
    }

  for (int i = 0; i != nleaf * 2 + 10; ++i)
    if (T[i].par == -1 && T[i].left != -1 && T[i].right != -1) rid = i;

  for (int i = 0; i != numfigs; ++i) free(figuration[i]);
  SetOneMut(figuration);
  //  for (int i = 0; i != numfigs; ++i) printf("%d\t%s\n", i, figuration[i]);

  return 1;
}



//--------------- <chr, pos> -----------
//     -1              0          1
//   <chri, posi>

int CTree::Compare(int chri, int posi)
{
  if (chri == chr && posi == pos) return 0;
  if (chri < chr || (chri == chr && posi < pos)) return -1; 
  return 1;
}



int CTree :: Bigger(int chri, int posi)
{
  if (chr > chri || (chr == chri && pos > posi)) return 1;
  return 0;
}

int CTree :: Smaller(int chri, int posi)
{
  if (chr < chri || (chr == chri && pos < posi)) return 1;
  return 0;
}



int CTree :: Equal(int chri, int posi)
{
  if (chr == chri && pos == posi) return 1;
  return 0;
}



int CTree::Equal(CTree *t)
{
  return Equal(t->GetChr(), t->GetPos());
}


//--------------------------------------------------------------------------------
void CTree :: InitLikelihoodData(int *rindex, double *d)
{
  index = rindex;
  for (int i =0; i != nsp; ++i)
    if (index[i] == -1) logprob[i] = NULL;
    else logprob[i] = &d[ index[i] * NG];//index [i]

  if (logConf == NULL) logConf = CallocDouble(numfigs, NA*NA); 

  for (int i =0; i != numfigs; ++i)
    for (int a = 0; a != NA; ++a)
      for (int b = 0; b != NA; ++b)
	if (a != b) logConf[i][a*NA + b] = LogProb(figuration[i], a, b);
	else logConf[i][a*NA+b] = -1e+10;
	  
  maxlogProb = logConf[0][0];
  for (int i =0; i != numfigs; ++i)
    for (int j =0; j != NA*NA; ++j)
      if (logConf[i][j] > maxlogProb) maxlogProb = logConf[i][j];

}



double CTree :: LogProb(char *fig, int a, int b)
{

  int order[] = {0,1,2,3,
		 1,4,5,6,
		 2,5,7,8,
		 3,6,8,9};
  
  int id_aa = order[a*NA+a];
  int id_ab = order[a*NA+b];
  int id_bb = order[b*NA+b];

  double sumlog  = 0;
  for (int i = 0; i != nsp; ++i)
    {
      double *p = logprob[ i  ];
      if (p == NULL) continue;

      if (fig[2*i] != fig[2*i+1])  sumlog += p[id_ab];
      else
	if (fig[2*i]  == 0 || fig[2*i] - '0' == 0) sumlog += p[id_aa];
	else sumlog += p[id_bb];
    }

  return sumlog;
}



