#ifndef __CTREE__
#define __CTREE__

#include "string.h"
#include "stdio.h"
#include "stdlib.h"
#include "Const.h"


struct Node{
  int left; //left child
  int right; //right child
  char leftdata;//if left child contains data
  char rightdata;  //if right child contains data
  int par;//parent
  int mm[2];
};

const long MaxT = 10000;

/*the idea
when we initialize a tree, 
we generate all possible configurations (stored in figuration) caused by mutations 
and calculate prior probaility of these configurations (store in PriorP)

when we put the data of samples (prob(data | g)), we calculate the probability 
log p(configuration | (aa), (ac), ..., tt) 
where (x,y) :: 0 = x, 1 = y
*/

class CTree
{
 private:
  Node *T; //node of tree
  int nleaf;//number of leaves
  int chr, pos; //location, [chr, pos]
  /*
                       nid = v
                      /       \
		     o         o
		   /  \       / \
                  v    v     v   v
  */

  void Put(int nid, char v, char *g); //set value v at node nid and all of its child

  int rid;
  void View(int id);//view trees 

  void View(int id, char** s); //view tree with labels s at leaves

  char *InitLeaves(void); //return (char*), all leaves = '0'

  int nsp; //number of samples, nsp = nleaf / 2
  int maxsp; //max samples

  char * figuration[MAXSAMPLE]; //figurations of trees with possible mutations. maximum MAXSAMPLE figuration
  double **logConf; //init log Conf
  //logConf[i][a*NA+b] = log Prob(Conf[i] | a,b): a = 0, b = 1


  double PriorP[MAXSAMPLE];     //prior Probability of figurations
  int numfigs;            //number of figurations
  void CalPriorProp(char** conf, int numfigs); //calculate prior Probability for configuration
  void ViewL(int id, char *s);


  double maxlogProb;
  double *logprob[MAXSAMPLE]; //log probability of data given genotypes
  //logorob[i]

  int *index;

  double LogProb(char* fig, int a, int b);
  
 public:
  double GetMaxLogProb(void) { return maxlogProb;}

  CTree();
  ~CTree();

  char**GetFig(void) {return figuration;} //Return figuration
  int GetNumGens(void) { return numfigs;} //return number of figurations
  int GetNumFigs(void) { return numfigs;} //return number of figurations
  double *GetPP(void) { return PriorP;}   //return prior probability
  double **GetLogProbFig(void) { return logConf;}
  
  int *GetIndex(void) { return index;}


  void SetOneMut(char **out);             //Set One Mutation: build figurations caused by one mutation  

  int Init(FILE *f); //input a tree from files


  //input a tree from a vector p, with nleaf at (rchr, pos)
  //p : (child1, child2, parent)
  int Init(int *p, int nleaf, int rchr, int rpos); 

  //-1 : smaller, 0: eq, 1 bigger
  int Compare(int chri, int posi);// { return Comp(chri, posi, chr, pos);}


  int Bigger(int chri, int posi); //(chr, pos) >(chri, posi)
  int Smaller(int chri, int posi);//(chr, pos) < (chri, posi)
  int Equal(int chri, int posi); //(chr, pos) == (chri, posi)

  int Equal(CTree *t);
  char GetChr(void) { return chr;} //get Chr
  int  GetPos(void) { return pos;}  //get Pos
  int  GetNumLeaf(void) { return nleaf;} //get number of leaves


  //  void InitLikelihoodData(int *index, double *d);
  //  void SetLikelihoodData(COneSiteLK *os); //set likelihood data for one site


  void ViewFig(void); //view figuration
  int  CheckFig(char *s);
  void ViewL(char *s);
  void View(void);//view tree
  void Display(void) { View();}

  void View(char **s); //view tree with labels at leaves from s



  double MaxProbLogProbFig(void);

};

#endif
