#include "SNPCandidate.h"
#include "CReadStream.h"
#include "CSNPDetection.h"
#include "Const.h"
#include "CCall.h"
#include "CReadData.h"
#include "COutput.h"
#include "CData4Phase.h"

#include "CPhase.h"


#define UNKNOWN 0
#define SNPCANDIDATE 1
#define SNPCALL 2

char NUC [] = "ACGT";

void AddName(int np, char **name)
{
  FILE *f = stdin;
  if (np > 3) f = fopen(name[3], "r");
  char bf[K10];
  while (!feof(f))
    {
      strcpy(bf, "");
      fgets(bf, K10, f);
      int len = strlen(bf) -1 ;
      while (len > 0 && bf[len] < 33) len --;
      bf[len]  = '\0';
      if (len > 1) printf("%s\t%s\n", bf, name[2]);
    }
}

void HelpCom(void)
{
  printf("./QCALL options\n");
  printf("\t-i data file| stdin\n");
  printf("\t-snpcan: filter by SNPCan\n");
  printf("\t-ct < low candidate threshold | 0.001>\n");
  printf("\t-co <file output for SNP candidates | - : stdoutmake> \n");
  printf("\t-vcf <vcf snp lists\n");
  printf("\t-sn <file of sample names> : output genotypes of samples given in the file\n");
  printf("\t-pphet <prior probability  of heterozygous |  .01>\n");
  printf("\t-ARG: <samples> <trees>  <bgzip>  with Margarita tree format, bgzip = 1: using bgzip file, 0 otherwises \n");
  printf("\t-farg: <tree> <samples> with fast Arg tree format \n");
  printf("\t-pho: output files of phased haplotypes using ARG trees\n");
  printf("\t-onego: genotyping/phasing with ARGs given as input. No new ARGs is built\n");
  printf("\t-indel: input as indel format\n");
  printf("\t-mns <n>: minimum number with data\n");
  printf("\t-filter <file name>: filtering sites that are not  in the given file\n");
  printf("\t-test: running for sites that do not include in ARGs\n");
  printf("\t-checktree: check tree distance to SNP candidate sites\n");
  printf("\t-r chr:pos1-pos2: running in this region\n");
}

int main(int ArgI, char*ArgC[])
{
  /*
  FILE *f = stdin;

  CReadDataOneLK *data = new CReadDataOneLK;
  data->Init(f);
  data->Test();
  //  data->Read();

  return 0;
  */
  if (ArgI == 1) { HelpCom(); return 0;}

  int Searchable = 0;
  FILE *fin = stdin;
  CReadData *data = new CReadData;
  

  CSNP *snpdec = NULL;
  //  CCall *call = NULL;
  CARG_m *arg_m = NULL;
  CARG_f* farg = NULL;

  COutput *outcan = NULL;// = new COutput;
  COutput *outphase = NULL;// = new outphase;
  
  int onego = 0;
  int minns = 0;
  int test = 0;
  int checktree = 0;

  int rpos1 = 0, rpos2 = 0, rchr = 0;

  int i = 0;
  for (; i != ArgI; ++i)
    {

      if (strcmp(ArgC[i], "-r") == 0) //region CHR:pos1-pos2
	{
	  char tempc[1024];
	  strcpy(tempc, ArgC[i+1]);
	  strstr(tempc, ":")[0] = ' ';
	  strstr(tempc, "-")[0] = ' ';
	  sscanf(tempc, "%d %d %d", &rchr, &rpos1, &rpos2);
	}

      if (strcmp(ArgC[i], "-i") == 0) 	  { fin = fopen(ArgC[i+1], "r"); Searchable = 1;}
      if (strcmp(ArgC[i], "-snpcan") == 0 && snpdec == NULL) snpdec = new CSNP;

      if (strcmp(ArgC[i], "-pphet") == 0) 
	{
	  if (outcan == NULL) outcan = new COutput;
	  outcan->SetPriorProbHom(1 - atof(ArgC[i+1]));
	}

      if (strcmp(ArgC[i], "-ct") == 0)
	{
	  if (snpdec == NULL) snpdec = new CSNP;
	  snpdec->SetThreshold(atof(ArgC[i+1]));
	}

      if (strcmp(ArgC[i], "-co") == 0) { 
	if (outcan == NULL) outcan = new COutput;
	outcan->Init(ArgC[i+1]);
      }

      if (strcmp(ArgC[i], "-pho") == 0) 
	{
	  if (outphase == NULL) outphase = new COutput;
	  outphase->Init(ArgC[i+1]);
	}

      if (strcmp(ArgC[i], "-sn") == 0) data->SetSampleNames(ArgC[i+1]);

      if (strcmp(ArgC[i], "-indel") == 0) data->SetIndel();

      if (strcmp(ArgC[i], "-mns") == 0) {
	minns = atoi(ArgC[i+1]);//data->SetMinData(atoi(ArgC[i+1]));
	printf("minns = %d\n", minns);
      }

      if (strcmp(ArgC[i], "-ARG") == 0) 
	{
	  if (i + 2 >= ArgI)
	    {
	      printf("error in -ARG option;\n");
	      exit(1);
	    }
	  if (arg_m == NULL) arg_m = new CARG_m;

	  arg_m->Init(ArgC[i+1], ArgC[i+2], atoi(ArgC[i+3]));

	  data->SetSampleNames(ArgC[i+1]);
	}

      if (strcmp(ArgC[i], "-onego") == 0) onego = 1;

      if (strcmp(ArgC[i], "-filter") == 0) data->SetFilter(ArgC[i+1]);
      if (strcmp(ArgC[i], "-test") == 0) test = 1;

      if (strcmp(ArgC[i], "-checktree") == 0) checktree = 1;

    }

  if (fin == NULL) 
    {
      printf("no input data\n");
      exit(1);
    }


  if (Searchable && rchr != 0)
    TextFileSeek(fin, rchr, rpos1);


  data->InitInputFile(fin);
  if (snpdec != NULL) snpdec->Init(400); 

  if (outcan != NULL) {
    outcan->InitSampleNames(data->GetNSample(), data->GetSampleNames());
    if (data->GetIndel()) outcan->SetIndel();
  }

  if (outphase != NULL) {
    outphase->InitSampleNames(data->GetNSample(), data->GetSampleNames());
    if (data->GetIndel()) outphase->SetIndel();                       
  }

  if (snpdec != NULL)
    outcan->VCF_Write_Header(0);//genotyping only

  if (outphase != NULL)    outphase->VCF_Write_Header(1);


  CData4Phase *dataphase = new CData4Phase;
  dataphase->InitSampleNames(data->GetNSample(), data->GetSampleNames());

  char *genotype = NULL;
  char *genconf = NULL;
  char *haplotype = NULL;
  char *hapconf = NULL;

  //  CHaplotypes *phasing = new CHaplotypes;

  if (outphase != NULL && data->GetNSample() > 0)
    {
      genotype = (char*)calloc(10*K10, sizeof(char));
      genconf = (char*)calloc(10*K10, sizeof(char));
      haplotype = (char*)calloc(10*K10, sizeof(char));
      hapconf = (char*)calloc(10*K10, sizeof(char));

      if (arg_m != NULL) {
	arg_m->InitLikelihoodName(data->GetSampleNames(), data->GetNSample());//order must be 0,1,2,etc.
	outphase->InitSampleNames(arg_m->GetNsample(), arg_m->GetNames());
      }
    }

  CPhased *phase = new CPhased;
  phase->InitSampleName(data->GetNSample(), data->GetSampleNames());

  int nsites = 0;
  char note[10240];

  while (1==1)
    {
      if (++nsites % 1000 == 0) fprintf(stderr, "nsites  = %d\n", nsites);

      COneSiteLK *os = data->Read();
      if (os == NULL) break; //end of data

      if (os->GetNS() < minns) continue;
      if (!os->CheckSNPConf()) continue;
      if (DISPLAY) os->Display();
      if (os->GetRef() == 'N') continue;


      int chr = os->GetChrInt();
      int pos = os->GetPos();
      int posend = os->GetPosEnd();

      if (DISPLAY) printf("chr = %d\t, pos = %d\n", chr, pos);

      if (rchr > 0)
	{
	  if (chr < rchr) continue;
	  if (chr > rchr) break;
	  if (pos < rpos1) continue;
	  if (pos > rpos2) break;
	}

      char alt = '.';
      double conf = 0, maf = 0;

      if (snpdec != NULL) 
	{
	  snpdec->GetSNPConfidence(os, conf, alt, maf);
	  os->SetConf(conf);
	  os->SetAlt(alt);
	  if (conf < snpdec->GetThreshold()) continue; //filter by SNP can
	  if (outcan != NULL) outcan->Output(os);
	}

      if (outphase == NULL) continue;//no output for  phase


      if (onego && arg_m != NULL) //phased with given ARGs
	{
	  unsigned char *phredlk = os->GetPhred();
	  CTree **T = (CTree**)calloc(100, sizeof(CTree*));;
	  int nt = arg_m->GetTrees(os->GetChr(), os->GetPos(), T);

	  if (DISPLAY) for (int i =0; i != nt; ++i) T[i]->Display();
	  //	  printf("got tree: %d\n", nt);

	  if (nt == 0)
	    {
	      printf("no tree at %d %d\n", chr, pos);
	      free(T);	      continue;
	    }


	  //	  os->Display();
	  printf("ntree = %d: %d %d\n", nt, chr, pos);

	  if (checktree)
	    {
	      int xx[100];
	      int nid = 0;
	      for (int i =0; i != nt; ++i)
		{
		  int t = T[i]->GetPos() - os->GetPos();
		  int skip = 0;
		  for (int i = 0; i != nid; ++i)
		    if (xx[i] == t) skip = 1;
		  if (skip) continue;
		  xx[nid++] = t;
		}
	  
	      printf("%s\t%d", os->GetChr(), os->GetPos());
	      for (int i =0; i != nid; ++i) printf("\t%d", xx[i]);
	      printf("\n");
	    }

	  if (test)
	    {
	      int skip = 0;
	      for (int i = 0; i != nt; ++i)
		if (T[i]->Equal(chr, pos)) skip = 1;

	      if (skip) 
		{
		  free(T);
		  continue; //running for sites that are not in ARGs
		}
	    }

	  //	  printf("geting here: L321\n");
	  for (int i = 0; i != nt; ++i) T[i]->InitLK(phredlk);
	  double conf = phase->ConfCal(T, nt, os->GetRefID());

	  //	  printf("geting here: L321\n");
	  os->SetConf(conf);

	  strcpy(note, "");
	  if (data->GetIndel())
	    {
	      if (T[0]->Smaller(chr, pos) && T[nt-1]->Bigger(chr, pos) && T[nt-1]->Smaller(chr, posend))  // T[0] < (chr, pos) < T[nt] < (chr, pos_end)
		{
		  sprintf(note, "Indel over SNPs: SNP = [%d:%d;%d:%d] and indel = [%d:%d-%d]", T[0]->GetChr(), T[0]->GetPos(), T[nt-1]->GetChr(), T[nt-1]->GetPos(), chr, pos, posend);
		  os->SetComments(note);
		  fprintf(stderr, "%s\n", note);
		}
	    }

	  if (conf > .01)// || 1==1) //10%
	    {
	      //	      printf("conf = %f\n", conf);
	      COneSite *s = new COneSite;
	      s->SetChrPos(os->GetChr(), os->GetPos());
	      s->InitLength(data->GetNSample() * 2);
	      s->SetConf(conf);

	      phase->Phasing(T, nt, os->GetRefID(), s);//get phasing data 
	      phase->Genotyping(T, nt, os->GetRefID(), s);//get phasing data 

	      if (outphase != NULL) outphase->OutPut(s, os);
	      delete s;//delete s
	    }

	  free(T);
	  continue;
	}

      dataphase->Add(os);//store data for phase
      data->NewOS();     //new one site data in read data
    }


  //  dataphase->Display();


  //  printf("starting ARG\n");
  //  exit(1);

  if (dataphase->GetNSite() != 0)
    {
      CARG_f* larg = phase->BuildHap_Phred(dataphase, arg_m, farg);
      CTree **T = (CTree**)calloc(100, sizeof(CTree*));;

      if (arg_m != NULL) arg_m->Reset();
      for (int i = 0; i != dataphase->GetNSite(); ++i)
	{
	  COneSiteLK* os = dataphase->GetSite(i);

	  int chr = os->GetChrInt();
	  int pos = os->GetPos();

	  unsigned char *phredlk = os->GetPhred();
	  //get tree from ARG_m, farg, larg
	  int ntree = 0;	  
	  if (arg_m != NULL) ntree += arg_m->GetTrees(chr, pos, &T[ntree]);
	  
	  if (test)
	    {
	      int skip = 0;
	      for (int i = 0; i != ntree; ++i)
		if (T[i]->Equal(chr, pos)) skip = 1;

	      if (skip) 
		{
		  printf("skip: %d\t%d\n", chr, pos);;
		  //		  free(T);
		  continue; //running for sites that are not in ARGs
		}
	    }

	  //	  ntree = 0;

	  if (farg != NULL)  ntree += farg ->GetTrees(chr, pos, &T[ntree]);
	  if (larg != NULL)  ntree += larg ->GetTrees(chr, pos, &T[ntree]);
	  for (int i = 0; i != ntree; ++i) T[i]->InitLK(phredlk);


	  double conf = phase->ConfCal(T, ntree, os->GetRefID());

	  COneSite *s = new COneSite;
	  s->SetChrPos(os->GetChr(), os->GetPos());
	  s->InitLength(data->GetNSample() * 2);
	  s->SetConf(conf);

	  phase->Phasing(T, ntree, os->GetRefID(), s);//get phasing data 
	  phase->Genotyping(T, ntree, os->GetRefID(), s);//get phasing data 
	  if (outphase != NULL) outphase->OutPut(s, os);
	}
      free(T);
      delete larg;
    }

  delete data;
  if (outcan != NULL) delete outcan;
  if (outphase != NULL) delete outphase;
  if (arg_m != NULL) delete arg_m;
  delete phase;
  delete dataphase;
  return 0;
}
