
#include "SNPCandidate.h"


 CSNP::CSNP()
 {
   maxn = 0;
   pp = NULL;
   Z = NULL;
   bf1 = NULL;
   bf2 = NULL;
   PPHom = NULL;
   PPHet = NULL;

   pphet = NULL;

   theta = 0.001; //theta for human
   threshold = 0.001;
   int gid = 0;
   for (int h1 = 0; h1 != NA; ++h1)
     for (int h2 = h1; h2 != NA; ++h2) 
       {
	 order[h1][h2] = gid;
	 order[h2][h1] = gid;
	 gid++;
       }

   if (DEBUG)
     for (int h1 = 0; h1 != NA; ++h1)
       {
	 for (int h2 = 0; h2 != NA; ++h2)  printf("%d\t", order[h1][h2]);
	 printf("\n");
       }
   


   
   //   char DNA[] = "ACGT";
   strcpy(DNA, "ACGT");
   for (int i = 0; i != K1; ++i)   OrderDNA[i] = -1;
   for (int i = 0; i != 4; ++i)    OrderDNA[ (int)  DNA[i]]  = i;
   for (int i = 0; i != 4; ++i)    OrderDNA[ (int) DNA[i] + 32]  = i;


  double log10 = log(10);
  for (int i = 0; i != K1; ++i)
    {
      double x = i;
      Phred2Prob[i] = exp(-x/10 *log10);
    }

 }



void CSNP :: Init(void) { Init(400);}

//initialize the number of samples
void CSNP::Init(int maxN)
{
  maxn = maxN + 1;

  //init pp
  //  PPhom: prior probability of hom
  //  PPhet: prior probability of het

  PPHet  = (double*)calloc(maxn, sizeof(double));
  PPHom  = (double*)calloc(maxn, sizeof(double));
  pphet  = (double**)calloc(maxn, sizeof(double));

  for (long n = 1 ; n != maxn; ++n)
    {
      double *p = (double*)calloc(2*n+10, sizeof(double));
      double sum = 0;

      for (long i = 1; i != 2*n; ++i)
	p[i] = (theta/2) * ( 1.0/i + 1.0/(2*n-i) );

      for (int i =1; i != 2*n;  ++i) sum += p[i];
      for (int i =1; i != 2*n;  ++i) p[i] /= sum;

      PPHet[n] = sum/6;
      PPHom[n] = (1-sum)/4;
      pphet[n] = p;
    }


   //Normalised factor for a paritioning
   Z = (double**)calloc(maxn, sizeof(double*));
   for (long i = 0; i != maxn; ++i)
     Z[i] = (double*)calloc((2*i+10), sizeof(double));


   Z[0][0] = 0; 
   Z[1][0] = 1;
   Z[1][1] = 1;
   Z[1][2] = 1; 

   for (long n = 2; n != maxn; ++n)
     for (long k = 0; k != 2*n+1; ++k)
       {
	 Z[n][k] = 0;
	 if (k <= 2*(n-1))   Z[n][k] += Z[n-1][k]; // r_n = bb
	 if (k >= 1) Z[n][k] += 2*Z[n-1][k-1]; //r_n = ab
	 if (k >= 2) Z[n][k] += Z[n-1][k-2]; //r_n = aa
       }


   //two buffers
   bf1 = (double*)calloc(2*maxn+10, sizeof(double));
   bf2 = (double*)calloc(2*maxn+10, sizeof(double));
 }

CSNP :: ~CSNP()
 {

   for (int i = 0; i != maxn; ++i) free(Z[i]);
   if (Z !=NULL) free(Z);


   if (PPHet != NULL) free(PPHet);
   if (PPHom != NULL) free(PPHom);

   for (int i = 1; i < maxn; ++i) free(pphet[i]);
   if (pphet != NULL) free(pphet);


   if (bf1 != NULL) free(bf1);
   if (bf2 != NULL) free(bf2);
 };


 //detect probability of SNPs from 
 //prob[ni*NG]
 //prob[0-->9] P[0]
 //prob[10-19] P[1]
 //....

 double* CSNP :: SNP(double *prob, int ni)
 {
   return SNP(prob, ni, -1);
 }



 double* CSNP :: SNP(double *prob, int ni, int rid)
 {
   n = ni;
   pp = prob;

   double pphom = PPHom[n];
   double pphet = PPHet[n];

   //normalized pp
   for (long i = 0; i != n; ++i)
     {
       double *p =&(pp[i*NG]);
       double maxp = 0;
       for (int j = 0; j != NG; ++j) 
	 if (p[j] > maxp) maxp = p[j];

       if (maxp > 0)
	 for (int j = 0; j != NG; ++j) p[j] /= maxp;
     }
   

   double *p = (double*)calloc(NG, sizeof(double));
   for (int i = 0; i != NG; ++i) p[i] = 0;


   if (rid != -1)
     {
       int gid = order[rid][rid];
       p[gid] = 1;
       int comp = 0;
       for (int i = 0; i != n; ++i)
	 if (prob[i*NG+gid] < 1.0) { comp = 1; break;}
       if (!comp) return p;
     }

   for (long h1 = 0; h1 != NA; ++h1)
     for (int h2 = h1; h2 != NA; ++h2)
       if (h1 <= h2)
	 {
	   int gid = order[h1][h2];
	   p[gid] = Computing(h1, h2);
	   if (h1 == h2) p[gid] *= pphom;
	   else p[gid] *= pphet;
	 }

   Norm(p);
   return p;
 }

 void CSNP :: Norm(double *p) { 
   Norm(p, NG);
 }

 void CSNP :: Norm(double *p, int n)
 {
   double sum = 0;
   for (int i =0; i != n; ++i)   sum += p[i];
   for (int i =0; i != n; ++i)   p[i] /= sum;
 }


 double CSNP :: ProbGen(long i, int h1, int h2)
 {
   int gid = order[h1][h2];
   return pp[i*NG + gid];
 }


 double CSNP :: Computing(int h1, int h2)
 {
   if (h1 == h2)
     {
       double out = 1;
       for (long i = 0; i != n; ++i)
	 {
	   double tempf = ProbGen(i, h1, h1);
	   out *= tempf;
	 }
       return out;
     }

   double *p1 = bf1;
   double *p2 = bf2;

   p1[0] = ProbGen(0, h2, h2); //p[r_0][0]
   p1[1] = ProbGen(0, h1, h2); //p[r_0][1]
   p1[2] = ProbGen(0, h1, h1); //p[r_0][2]

   p1[3] = 0;
   p1[4] = 0;

   for (long i = 1; i != n; ++i)
     {
       int l = 2*(i + 1); //l = i+1, number of r_0, ..., r_{i}

       if (i % 2 == 1) { p1 = bf1 ; p2 = bf2;}
       else {p1 = bf2; p2 = bf1;}

       double paa = ProbGen(i, h1, h1);
       double pab = ProbGen(i, h1, h2);
       double pbb = ProbGen(i, h2, h2);

       if (paa == 0 && pab == 0 && pbb == 0)
	 {
	   paa = 1;
	   pab = 1;
	   pbb = 1;
	 }

       //p1: 0-->2i

       p2[0] = p1[0] * pbb;
       p2[1] = p1[1] * pbb + p1[0] * pab;

       for (long k = 2;  k != l+1; ++k)
	 p2[k] = p1[k] * pbb + 2*p1[k-1] * pab + p1[k-2] * paa;

       p2[l+1] = 0;
       p2[l+2] = 0;
     }

   double *p;
   if ((n -1) % 2 == 1) p = bf2;
   else p = bf1;


   double out = 0;
   for (long i = 1 ; i!= 2*n; ++i)
     out += (pphet[n][i] * p[i]) / Z[n][i];
   return out;
 }



 double CSNP :: SNPConfidence (double *prob, int ns, int rid)
 {
   exit(1);
   if (rid == -1 || rid > 4)
     {
       printf("error: %d\n", rid);
       exit(1);
     }

   double *p = SNP(prob, ns, rid);
   double out = 1 - p[ order[rid][rid]];
   free(p);
   return out;
 }






void CSNP :: SNPConfidence (double *prob, int ns, int &rid, double &conf, char &altall, double &maf)
{
  conf = 0;
  altall = '.';
  maf = -1;


  //  if (ns < 3) return;
  if (ns < 2) return;

  double *p = SNP(prob, ns, rid);
  //  Display(p, NG);

  double maxp = 0;

  if (rid != -1)
    {
      conf = 1 - p[ order[rid][rid]];
      for (int i = 0; i != NA; ++i)
	if (p[order[i] [rid]] > maxp && rid != i) {
	  altall = DNA[i]; 
	  maxp = p[order[i][rid]];
	}
    }
  else
    {
      conf = 0;
      rid = 0;
      for (int a = 0; a != NA; ++a)
	for (int b = a+1; b != NA; ++b)
	  {
	    conf +=  p [order[a][b]];
	    if ( p [order[a][b]] > maxp) {
	      altall = DNA[a];
	      rid = b;
	      maxp = p[order[a][b]];
	    }
	  }
    }
  free(p);
}

void CSNP::GetSNPConfidence(COneSiteLK *os, double &conf, char &alt, double &maf)
{
  int ns = os->GetND();
  double *prob = (double*)calloc(ns * NG, sizeof(double));
  unsigned char *p = os->GetPhred();
  int *depth = os->GetDepth();
  int nd = 0;
  for (int i = 0; i != ns; ++i)
    if (depth[i] > 0) 
      {
	for (int j = 0; j != NG; ++j)
	  prob[nd*NG + j] = Phred2Prob[ p[i*NG + j] ];
	nd++;
      }

  int rid = OrderDNA[ (int) os->GetRef()];
  if (DEBUG)
    for (int i =0; i != ns*NG; ++i) {
      if (i % NG == 0) printf("\n");
      printf("%f\t", prob[i]);
    }

  SNPConfidence(prob, nd, rid, conf, alt, maf);
  //  printf("%d\t%c\n", rid, alt);
  if (os->GetRef() == 'N') os->SetMissingRef(DNA[rid]);
  free(prob);
}
